#!/usr/bin/env python3

"""
rc_eval.py
Standalone script to evaluate RC values given an aimless shooting working directory and reaction coordinate definition
"""

import sys
import os
import glob
import argparse
import pickle
import time
import math
from atesa import utilities

def update_progress(progress, message='Progress', eta=0, quiet=False):
    """
    Print a dynamic progress bar to stdout.

    Credit to Brian Khuu from stackoverflow, https://stackoverflow.com/questions/3160699/python-progress-bar

    Parameters
    ----------
    progress : float
        A number between 0 and 1 indicating the fractional completeness of the bar. A value under 0 represents a 'halt'.
        A value at 1 or bigger represents 100%.
    message : str
        The string to precede the progress bar (so as to indicate what is progressing)
    eta : int
        Number of seconds to display as estimated completion time (converted into HH:MM:SS)
    quiet : bool
        If True, suppresses output entirely

    Returns
    -------
    None

    """

    if quiet:
        return None

    barLength = 10  # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done!          \r\n"
    block = int(round(barLength * progress))
    if eta:
        # eta is in seconds; convert into HH:MM:SS
        eta_h = str(math.floor(eta/3600))
        eta_m = str(math.floor((eta % 3600) / 60))
        eta_s = str(math.floor((eta % 3600) % 60)) + ' '
        if len(eta_m) == 1:
            eta_m = '0' + eta_m
        if len(eta_s) == 2:
            eta_s = '0' + eta_s
        eta_str = eta_h + ':' + eta_m + ':' + eta_s
        text = "\r" + message + ": [{0}] {1}% {2}".format("#" * block + "-" * (barLength - block), round(progress * 100, 2), status) + " ETA: " + eta_str
    else:
        text = "\r" + message + ": [{0}] {1}% {2}".format("#" * block + "-" * (barLength - block), round(progress * 100, 2), status)
    sys.stdout.write(text)
    sys.stdout.flush()


def main(working_directory, rc_definition, as_out_file, extrema=False):
    """
    The main function of rc_eval.py. Accepts an aimless shooting working directory and a reaction coordinate definition,
    producing in that directory a new file named 'rc.out' (overwriting if one already exists) identifying each shooting
    point (files in the directory whose names end in "_init.rst7") and its corresponding reaction coordinate value in a
    sorted list.

    If extrema == True, skips producing rc.out and just returns the minimum and maximum RC value for a single accepted
    shooting move, which is useful when preparing umbrella sampling simulations.

    Parameters
    ----------
    working_directory : str
        The path to the aimless shooting working directory in which to act
    rc_definition : str
        A reaction coordinate definition formatted as a string of python-readable code with "CV[X]" standing in for the
        Xth CV value (one-indexed); this RC definition should be in terms of reduced variables (values between 0 and 1)
    as_out_file : str
        Path to the aimless shooting output file used to build the reaction coordinate. Usually this should be a
        decorrelated file (named with "decorr").
    extrema : bool
        If True, skips producing rc.out and just returns the minimum and maximum RC value for a single accepted shooting
        move, which is useful when preparing umbrella sampling simulations.

    Returns
    -------
    None

    """

    # Change to working directory
    os.chdir(working_directory)

    # Unpickle settings object for use in utilities.get_cvs
    try:
        settings = pickle.load(open('settings.pkl', 'rb'))
    except FileNotFoundError:   # replace with more informative error message
        raise FileNotFoundError('the working directory must contain a valid settings.pkl file, which is generated '
                                'automatically when running ATESA, but one was not found in the working directory: '
                                + working_directory)

    if not settings.job_type == 'aimless_shooting':
        raise RuntimeError('rc_eval.py can only be called on an aimless shooting working directory, but the provided '
                           'directory (' + working_directory + ') contains a settings.pkl file with job_type = ' +
                           settings.job_type)

    settings.as_out_file = as_out_file      # for reducing CVs properly
    settings.include_qdot = False           # unnecessary for our purposes

    if extrema:
        from atesa.main import Thread
        print('Evaluating final RC values of forward and backward trajectories from an accepted shooting move...')
        result = []
        allthreads = pickle.load(open('restart.pkl', 'rb'))
        for thread in allthreads:
            if thread.history.last_accepted > -1:   # if accepted move exists in thread
                for job_index in range(2):
                    cvs = utilities.get_cvs(thread.history.prod_trajs[thread.history.last_accepted][job_index], settings, reduce=True, frame=-1).split(' ')
                    result.append(utilities.evaluate_rc(rc_definition, cvs))
                print(' Shooting move name: ' + thread.history.init_coords[thread.history.last_accepted][0])
                print(' extrema: ' + str([float('%.4f' % float(item)) for item in result]))
                return [float('%.4f' % float(item)) for item in result]   # to exit the script after returning extrema
        raise RuntimeError('none of the shooting moves in the working directory appear to contain any accepted moves.')

    # Obtain list of shooting point coordinate files
    file_list = glob.glob('*_init.rst7')
    if not file_list:
        raise FileNotFoundError('no valid shooting point files (as given by names ending in \'_init.rst7\') were found '
                                'in the working directory \'' + working_directory + '\'. Is this an aimless shooting '
                                'working directory?')

    # Iterate through the list, calling evaluate_rc for each one and storing the result
    results = []
    count = 0
    count_to = len(file_list)
    update_progress(0, 'Evaluating RC values')
    speed_data = [0, 0]
    for file in file_list:
        t = time.time()
        cv_list = utilities.get_cvs(file, settings, reduce=True).split(' ')
        results.append([file + ': ', utilities.evaluate_rc(rc_definition, cv_list)])
        this_speed = time.time() - t
        speed_data = [(speed_data[1] * speed_data[0] + this_speed) / (speed_data[1] + 1), speed_data[1] + 1]
        count += 1
        eta = (count_to - count) * speed_data[0]
        update_progress(count/count_to, 'Evaluating RC values', eta=eta)
    results = sorted(results, key=lambda x: abs(float(x[1])))  # sort results by absolute value of RC

    # Create and write to rc.out file
    open('rc.out', 'w').close()
    with open('rc.out', 'a') as f:
        for result in results:
            f.write(result[0] + str(result[1]) + '\n')
        f.close()


if __name__ == '__main__':
    if len(sys.argv) == 5:
        extrema = sys.argv[4]
    else:
        extrema = False
    if len(sys.argv) <= 3:
        raise RuntimeError('not enough arguments; you gave ' + str(len(sys.argv) - 1) + ' but rc_eval.py takes 3 or 4')
    main(sys.argv[1], sys.argv[2], sys.argv[3], extrema)



"""
This portion of the program is responsible for handling setup of the appropriate batch script(s) for the next step in a
Thread, passing them to a task manager to submit them, and updating the list of currently running threads accordingly.
"""

import django.template
import os
import sys
import warnings
from atesa import factory
from django.conf import settings as django_settings

def process(thread, running, settings):
    """
    The main function of process.py. Reads the thread to identify the next step, then builds and submits the batch
    file(s) as needed.

    Parameters
    ----------
    thread : Thread()
        The Thread object on which to act
    running : list
        The list of currently running threads, which will be updated as needed
    settings : argparse.Namespace
        Settings namespace object

    Returns
    -------
    running : list
        The updated list of currently running threads after this process step

    """

    if not django_settings.configured:  # need to configure just once per thread
        django_settings.configure()

    ### Determine next step and, if appropriate, build corresponding list of batch files ###
    if not thread.skip_update:
        thread.current_type, thread.current_name = thread.get_next_step(settings)
    else:
        thread.skip_update = False

    if thread.terminated:
        if thread in running:
            running.remove(thread)
            if running is None:
                running = []       # to keep running as a list, even if empty
        return running

    batchfiles = []         # initialize list of batch files to fill out
    jobtype = factory.jobtype_factory(settings.job_type)    # get jobtype for calling jobtype.update_history
    this_inpcrd = jobtype.get_inpcrd(thread)
    for job_index in range(len(thread.current_type)):
        type = thread.current_type[job_index]
        name = thread.current_name[job_index]
        inp = jobtype.get_input_file(thread, job_index, settings)
        template = settings.env.get_template(thread.get_batch_template(type, settings))

        these_kwargs = { 'name': thread.name + '_' + name,
                         'nodes': eval('settings.' + type + '_nodes'),
                         'taskspernode': eval('settings.' + type + '_ppn'),
                         'walltime': eval('settings.' + type + '_walltime'),
                         'mem': eval('settings.' + type + '_mem'),
                         'solver': eval('settings.' + type + '_solver'),
                         'inp': inp,
                         'out': thread.name + '_' + name + '.out',
                         'prmtop': thread.topology,
                         'inpcrd': this_inpcrd[job_index],
                         'rst': thread.name + '_' + name + '.rst7',
                         'nc': thread.name + '_' + name + '.nc',
                         'working_directory': settings.working_directory,
                         'extra': eval('settings.' + type + '_extra') }

        filled = template.render(django.template.Context(these_kwargs))
        newfilename = thread.name + '_' + name + '.' + settings.batch_system
        try:
            with open(newfilename, 'w') as newfile:
                newfile.write(filled)
                newfile.close()
        except OSError as e:
            if 'name too long' in str(e):
                warnings.warn('Encountered too-long filename: ' + newfilename + '. Skipping submission of this thread '
                              'to the task manager, terminating it, and continuing. Consider renaming a sample of the '
                              'initial coordinate files you like best and starting a new ATESA run. If this limitation '
                              'is a problem for you, please raise an issue on GitHub detailing your use-case.')
                thread.terminated = True
                return running

        batchfiles.append(newfilename)
        jobtype.update_history(thread, settings, **these_kwargs)

    ### Submit batch files to task manager ###
    taskmanager = factory.taskmanager_factory(settings.task_manager)
    thread.jobids = []      # to clear out previous jobids if any exist
    for file in batchfiles:
        thread.jobids.append(taskmanager.submit_batch(file, settings))

    if thread not in running:
        running.append(thread)
    return running



"""
Utility functions that don't properly fit anywhere else. These are operations that aren't specific to any particular
interface or script.
"""

import pytraj
import mdtraj
import shutil
import fileinput
import re
import os
import sys
import math
import numpy
import pickle
import copy
import subprocess
import warnings
import itertools
from multiprocess import Pool
from statsmodels.tsa import stattools

def check_commit(filename, settings):
    """
    Check commitment of coordinate file to basins defined by settings.commit_fwd and settings.commit_bwd.

    Raises a RuntimeError if the supplied file is not suitable for checking for commitment.

    Parameters
    ----------
    filename : str
        Name of coordinate or trajectory file to be checked. If the file has more than one frame, only the last frame
        will be loaded.
    settings : argparse.Namespace
        Settings namespace object

    Returns
    -------
    commit_flag : str
        Either 'fwd' or 'bwd' if the coordinates are in the corresponding basin, or '' if in neither

    """

    # todo: consider adding support for angles and dihedrals

    try:
        traj = mdtraj.load(filename, top=settings.topology)[-1]
        # traj = pytraj.load(filename, settings.topology, frame_indices=[-1])
        if traj.n_frames == 0:
            raise RuntimeError('Attempted to check commitment for file: ' + filename + ' but it has no frames.')
    except ValueError as e:
        raise RuntimeError('Unable to load file: ' + filename + ' for checking commitment with the specified topology '
                           'file.\npytraj returned error: ' + str(e))
    commit_flag = ''    # initialize

    for i in range(len(settings.commit_fwd[2])):
        if settings.commit_fwd[3][i] == 'lt':
            if mdtraj.compute_distances(traj, numpy.array([[settings.commit_fwd[0][i] - 1, settings.commit_fwd[1][i] - 1]]))[0][0] * 10 <= settings.commit_fwd[2][i]:
            # if pytraj.distance(traj, '@' + str(settings.commit_fwd[0][i]) + ' @' + str(settings.commit_fwd[1][i]),
            #                    n_frames=1)[0] <= settings.commit_fwd[2][i]:
                commit_flag = 'fwd'     # if a committor test is passed, testing moves on to the next one.
            else:
                commit_flag = ''
                break                   # if a committor test is not passed, all testing in this direction fails
        elif settings.commit_fwd[3][i] == 'gt':
            if mdtraj.compute_distances(traj, numpy.array([[settings.commit_fwd[0][i] - 1, settings.commit_fwd[1][i] - 1]]))[0][0] * 10 >= settings.commit_fwd[2][i]:
            # if pytraj.distance(traj, '@' + str(settings.commit_fwd[0][i]) + ' @' + str(settings.commit_fwd[1][i]),
            #                    n_frames=1)[0] >= settings.commit_fwd[2][i]:
                commit_flag = 'fwd'
            else:
                commit_flag = ''
                break
        else:
            raise ValueError('An incorrect committor definition \"' + settings.commit_fwd[3][i] + '\" was given for '
                             'index ' + str(i) + ' in the \'fwd\' direction.')

    if commit_flag == '':               # only bother checking for bwd commitment if not fwd committed
        for i in range(len(settings.commit_bwd[2])):
            if settings.commit_bwd[3][i] == 'lt':
                if mdtraj.compute_distances(traj, numpy.array([[settings.commit_bwd[0][i] - 1, settings.commit_bwd[1][i] - 1]]))[0][0] * 10 <= settings.commit_bwd[2][i]:
                # if pytraj.distance(traj, '@' + str(settings.commit_bwd[0][i]) + ' @' + str(settings.commit_bwd[1][i]),
                #                    n_frames=1)[0] <= settings.commit_bwd[2][i]:
                    commit_flag = 'bwd' # if a committor test is passed, testing moves on to the next one.
                else:
                    commit_flag = ''
                    break               # if a committor test is not passed, all testing in this direction fails
            elif settings.commit_bwd[3][i] == 'gt':
                if mdtraj.compute_distances(traj, numpy.array([[settings.commit_bwd[0][i] - 1, settings.commit_bwd[1][i] - 1]]))[0][0] * 10 >= settings.commit_bwd[2][i]:
                # if pytraj.distance(traj, '@' + str(settings.commit_bwd[0][i]) + ' @' + str(settings.commit_bwd[1][i]),
                #                    n_frames=1)[0] >= settings.commit_bwd[2][i]:
                    commit_flag = 'bwd'
                else:
                    commit_flag = ''
                    break
            else:
                raise ValueError('An incorrect committor definition \"' + settings.commit_bwd[3][i] + '\" was given for'
                                 ' index ' + str(i) + ' in the \'bwd\' direction.')

    return commit_flag


def get_cvs(filename, settings, reduce=False, frame=0, only_in_rc=False):
    """
    Get CV values for a coordinate file given by filename, as well as rates of change if settings.include_qdot = True.

    If reduce = True, the returned CVs will be reduced to between 0 and 1 based on the minimum and maximum values of
    that CV in as.out, which is assumed to exist at settings.as_out_file.

    If frame is > 0 or 'all', filename will be interpreted as a trajectory file instead of a .rst7 coordinate file.
    This option is incompatible with settings.include_qdot = True.

    Parameters
    ----------
    filename : str
        Name of .rst7-formatted coordinate file to be checked (or trajectory file if frame > 0 or frame == 'all')
    settings : argparse.Namespace
        Settings namespace object
    reduce : bool
        Boolean determining whether to reduce the CV values for use in evaluating an RC value that uses reduced values
    frame : int or str
        Frame of trajectory filename to use (0-indexed, negatives read from end); or, 'all' returns the CVs for each
        frame of the input trajectory, separated by newlines. 'all' is incompatible with settings.include_qdot = True
    only_in_rc : bool
        If this is True, then do not evaluate and return 0 for all CVs not present in the RC defined by
        settings.rc_definition. This is useful to speed up evaluation when all CVs are not necessary. Applies also to
        corresponding qdot values as appropriate.

    Returns
    -------
    output : str
        Space-separated list of CV values for the given coordinate file

    """

    def increment_coords():
        # Produces a new coordinate file from the existing one by incrementing coordinates by their rates of change
        # Returns the name of the newly-created coordinate file
        byline = open(filename).readlines()
        pattern = re.compile('-*[0-9.]+')           # regex to match numbers including decimals and negatives
        n_atoms = pattern.findall(byline[1])[0]     # number of atoms indicated on second line of .rst file

        shutil.copyfile(filename, filename + '_temp.rst7')
        for i, line in enumerate(fileinput.input(filename + '_temp.rst7', inplace=1)):
            if int(n_atoms)/2 + 2 > i >= 2:
                newline = line
                coords = pattern.findall(newline)                                          # line of coordinates
                try:
                    vels = pattern.findall(byline[i + int(math.ceil(int(n_atoms)/2))])     # corresponding velocities
                except IndexError:
                    os.remove(filename + '_temp.rst7')      # to clean up
                    os.remove(filename + '_temp.rst7.bak')  # to clean up
                    fileinput.close()
                    raise IndexError('get_cvs.increment_coords() encountered an IndexError. This is caused '
                             'by attempting to read qdot values from a coordinate file lacking velocity information, or'
                             ' else by that file being truncated. Ensure that the relevant simulation input file is set'
                             ' to write velocities to output trajectories, as this may not be default behavior. In '
                             'Amber, this is accomplished by setting ntwv=-1 in the MD input file. Alternatively, if '
                             'you don\'t want to include velocity terms in the output file, you can set include_qdot = '
                             'False in the ATESA config file.\n'
                             'The offending file is: ' + filename)

                # Sometimes items in coords or vels 'stick together' at a negative sign (e.g., '-1.8091748-112.6420521')
                # This next loop is just to split them up
                for index in range(len(coords)):
                    length = len(coords[index])                     # length of string representing this coordinate
                    replace_string = str(float(coords[index]) + float(vels[index]))[0:length-1]
                    while len(replace_string) < length:
                        replace_string += '0'
                    newline = newline.replace(coords[index], replace_string)
                sys.stdout.write(newline)
            else:
                sys.stdout.write(line)

        return filename + '_temp.rst7'

    def reduce_cv(unreduced_value, local_index, rc_minmax):
        # Returns a reduced value for a CV given an unreduced value and the index within as.out corresponding to that CV
        this_min = rc_minmax[0][local_index]
        this_max = rc_minmax[1][local_index]
        if this_min == this_max:
            return unreduced_value      # can't reduce when a CV only has a single value across rc_minmax
        return (float(unreduced_value) - this_min) / (this_max - this_min)

    def cv_is_in_rc(cv_index):
        # Helper function checks if the zero-indexed cv_index corresponds to a CV in settings.rc_definition
        if settings.rc_definition == '':
            raise RuntimeError('utilities.get_cvs was called with only_in_rc=True, but rc_definition was empty. You '
                               'must provide the RC definition in the configuration file to use this option.')
        # Have to be careful here not to interpret, e.g., "CV10" as matching "CV1", so we need to be a bit clever
        pattern = re.compile('CV[0-9]+')
        cvs_in_rc = pattern.findall(settings.rc_definition)
        cv_to_look_for = 'CV' + str(int(cv_index) + 1)   # cv_index is zero-indexed, but CVs in RCs are 1-indexed
        return cv_to_look_for in cvs_in_rc

    if frame == 'all' and settings.include_qdot:
        raise RuntimeError('utilities.get_cvs cannot be called with frame=all if settings.include_qdot == True')

    if not os.getcwd() == settings.working_directory:
        os.chdir(settings.working_directory)    # make sure we're in the working directory

    # boolean determining whether we need to bother loading pytraj
    need_pytraj = any([('traj' in cv and not 'mtraj' in cv) for cv in settings.cvs])

    # define variables as they are expected to be interpreted in CV definitions (these are potentially invoked by eval)
    delete_traj_name = False  # so we don't delete the traj_name file later unless we created it below
    if frame in [0, 'all']:
        full_mtraj = mdtraj.load(filename, top=settings.topology)
        if need_pytraj:  # only load pytraj if it's needed
            full_traj = pytraj.iterload(filename, settings.topology)
        traj_name = filename
    else:
        if need_pytraj:
            if frame < 0:
                full_traj = pytraj.iterload(filename, settings.topology)
                pframe = full_traj.n_frames + frame
                full_traj = pytraj.iterload(filename, settings.topology, frame_slice=[(pframe, pframe + 1)])
            else:
                full_traj = pytraj.iterload(filename, settings.topology, frame_slice=[(frame, frame + 1)])
        full_mtraj = mdtraj.load(filename, top=settings.topology)[frame]
        if True in ['traj_name' in cv for cv in settings.cvs]:  # if True, need .rst7 formatted file to operate on
            mdengine = factory.mdengine_factory(settings.md_engine)
            traj_name = mdengine.get_frame(filename, frame, settings)
            delete_traj_name = True     # to be sure we clean up traj_name later

    rc_minmax = [[],[]]
    if reduce:
        # Try to load from file if available
        size = os.path.getsize(settings.as_out_file)
        if os.path.exists(str(size) + '_minmax.pkl'):
            rc_minmax = pickle.load(open(str(size) + '_minmax.pkl', 'rb'))
        else:
            # Prepare new cv_minmax list
            asout_lines = [[float(item) for item in line.replace('A <- ', '').replace('B <- ', '').replace(' \n', '').replace('\n', '').split(' ')] for line in open(settings.as_out_file, 'r').readlines()]
            open(settings.as_out_file, 'r').close()
            mapped = list(map(list, zip(*asout_lines)))
            rc_minmax = [[numpy.min(item) for item in mapped], [numpy.max(item) for item in mapped]]
            pickle.dump(rc_minmax, open(str(size) + '_minmax.pkl', 'wb'), protocol=2)

    output = ''
    values = []
    sigfigs = '%.' + str(settings.sigfigs) + 'f'

    if frame == 'all':
        n_iter = full_mtraj.n_frames
    else:
        n_iter = 1

    for iter in range(n_iter):  # iterate over all frames
        local_index = -1
        # set traj and mtraj to only the desired frame; has no (important) effect if there's only one frame anyway
        mtraj = full_mtraj[iter]
        if need_pytraj:     # only load traj if it's needed
            traj = full_traj[iter:iter + 1:1]
        for cv in settings.cvs:
            local_index += 1
            if (not only_in_rc) or cv_is_in_rc(local_index):
                evaluation = eval(cv)       # evaluate the CV definition code (potentially using traj, mtraj, and/or traj_name)
                if settings.include_qdot:   # want to save values for later
                    values.append(float(evaluation))
                if reduce:                  # reduce values if necessary
                    evaluation = reduce_cv(evaluation, local_index, rc_minmax)
            else:
                evaluation = 0
            output += sigfigs % float(evaluation) + ' '
        if n_iter > 1:
            output += '\n'
        elif settings.include_qdot and not settings.job_type == 'equilibrium_path_sampling':  # if True, then we want to include rate of change for every CV, too
            # Strategy here is to write a new temporary .rst7 file by incrementing all the coordinate values by their
            # corresponding velocity values, load it as a new iterload object, and then rerun our analysis on that.
            incremented_filename = increment_coords()
            mtraj = mdtraj.load(incremented_filename, top=settings.topology)
            if need_pytraj:  # only load traj if it's needed
                traj = pytraj.iterload(incremented_filename, settings.topology)
            local_index = -1
            for cv in settings.cvs:
                local_index += 1
                if (not only_in_rc) or cv_is_in_rc(local_index):
                    evaluation = eval(cv) - values[local_index]  # Subtract value 1/20.455 ps earlier from value of cv
                    if reduce:
                        try:
                            evaluation = reduce_cv(evaluation, local_index + len(settings.cvs), rc_minmax)
                        except IndexError:
                            raise RuntimeError('attempted to obtain a reduced value for the rate-of-change ("qdot") of CV' +
                                               str(local_index) + ', but the corresponding column appears to be missing from '
                                               'the aimless shooting output file: ' + settings.as_out_file + '. If this is the '
                                               'wrong file, please specify the right one with the "as_out_file" option in the '
                                               'configuration file. If this is the right file but it does not contain qdot '
                                               'terms, please add "include_qdot = False" to the configuration file.')
                else:
                    evaluation = 0
                output += sigfigs % float(evaluation) + ' '
            os.remove(incremented_filename)     # clean up temporary file

    while output[-1] in [' ', '\n']:
        output = output[:-1]                # remove trailing space and/or newline on terminating line
    output = output.replace(' \n', '\n')    # remove trailing spaces on non-terminating lines

    if delete_traj_name:
        os.remove(traj_name)

    return output


def rev_vels(restart_file):
    """
    Reverse all the velocity terms in a restart file and return the name of the new, 'reversed' file.

    Parameters
    ----------
    restart_file : str
        Filename of the 'fwd' restart file, in .rst7 format

    Returns
    -------
    reversed_file : str
        Filename of the newly written 'bwd' restart file, in .rst7 format

    """

    byline = open(restart_file).readlines()
    open(restart_file).close()
    pattern = re.compile(r'[-0-9.]+')            # regex to match numbers including decimals and negatives
    pattern2 = re.compile(r'\s[-0-9.]+')         # regex to match numbers including decimals and negatives, with one space in front
    try:
        n_atoms = pattern.findall(byline[1])[0]     # number of atoms indicated on second line of .rst7 file
    except IndexError:
        raise RuntimeError('restart file: ' + restart_file + ' exists but is improperly formatted. It must contain the '
                           'number of atoms as the first integer on its second line.')
    offset = 2                  # appropriate for n_atoms is odd; offset helps avoid modifying the box line
    if int(n_atoms) % 2 == 0:   # if n_atoms is even...
        offset = 1              # appropriate for n_atoms is even

    try:
        name = restart_file[:restart_file.rindex('.')]  # everything before last '.', to remove file extension
    except ValueError:
        name = restart_file     # if no '.' in the filename

    shutil.copyfile(restart_file, name + '_bwd.rst7')
    for i, line in enumerate(fileinput.input(name + '_bwd.rst7', inplace=1)):
        if int(n_atoms) / 2 + 2 <= i <= int(n_atoms) + offset:  # if this line is a velocity line
            newline = line
            for vel in pattern2.findall(newline):
                if '-' in vel:
                    newline = newline.replace(vel, '  ' + vel[2:], 1)   # replace ' -magnitude' with '  magnitude'
                else:
                    newline = newline.replace(vel, '-' + vel[1:], 1)    # replace ' magnitude' with '-magnitude'
            sys.stdout.write(newline)
        else:  # if not a velocity line
            sys.stdout.write(line)

    return name + '_bwd.rst7'


def evaluate_rc(rc_definition, cv_list):
    """
    Evaluate the RC value given by RC definition for the given list of CV values given by cv_list.

    Parameters
    ----------
    rc_definition : str
        A reaction coordinate definition formatted as a string of python-readable code with "CV[X]" standing in for the
        Xth CV value (zero-indexed); e.g., "CV2" has value "4" in the cv_list [1, -2, 4, 6]
    cv_list : list
        A list of CV values whose indices correspond to the desired values in rc_definition

    Returns
    -------
    rc_value : float
        The value of the reaction coordinate given the values in cv_list

    """

    # Fill in CV[X] slots with corresponding values from cv_list
    for i in reversed(range(len(cv_list))):     # reversed so that e.g. CV10 doesn't get interpreted as '[CV1]0'
        rc_definition = rc_definition.replace('CV' + str(i + 1), str(cv_list[i]))

    # Evaluate the filled-in rc_definition and return the result
    return eval(rc_definition)


def resample(settings, partial=False, full_cvs=False, only_full_cvs=False):
    """
    Resample each shooting point in each thread with different CV definitions to produce new output files with extant
    aimless shooting data.

    This function also assesses decorrelation times and produces one or more decorrelated output files. If and only if
    settings.information_error_checking == True, decorrelated files are produced at each settings.information_error_freq
    increment. In this case, if partial == True, decorrelation will only be assessed for data lengths absent from the
    info_err.out file in the working directory.

    Parameters
    ----------
    settings : argparse.Namespace
        Settings namespace object
    partial : bool
        If True, reads the info_err.out file and only builds new decorrelated output files where the corresponding lines
        are missing from that file. If partial == False, decorrelation is assessed for every valid data length. Has no
        effect if not settings.information_error_checking.
    full_cvs : bool
        If True, also resamples as_full_cvs.out using every prod trajectory in the working directory.

    Returns
    -------
    None

    """

    # todo: test this more thoroughly using a dummy thread and a manual decorrelation time calculation using different software

    # This function is sometimes called from outside the working directory, so make sure we're there
    os.chdir(settings.working_directory)

    # Load in allthreads from restart.pkl
    try:
        allthreads = pickle.load(open('restart.pkl', 'rb'))
    except FileNotFoundError:
        raise FileNotFoundError('resample = True requires restart.pkl, but could not find one in working directory: '
                                + settings.working_directory)

    if not only_full_cvs:
        # Remove pre-existing output files if any, initialize new one
        open(settings.working_directory + '/as_raw_resample.out', 'w').close()
        if settings.information_error_checking:
            open(settings.working_directory + '/as_raw_timestamped.out', 'w').close()

        # Open files for writing outside loop (much faster than opening/closing for each write)
        f1 = open(settings.working_directory + '/as_raw_resample.out', 'a')
        if settings.information_error_checking:
            f2 = open(settings.working_directory + '/as_raw_timestamped.out', 'a')

        # Iterate through each thread's history.init_coords list and obtain CV values as needed
        for thread in allthreads:
            thread.this_cvs_list = []       # initialize full nested list of CV values for this thread
            thread.cvs_for_later = []       # need this one with empty lists for failed moves, for indexing reasons
            for step_index in range(len(thread.history.prod_results)):
                if thread.history.prod_results[step_index][0] in ['fwd', 'bwd']:
                    if thread.history.prod_results[step_index][0] == 'fwd':
                        this_basin = 'B'
                    else:  # 'bwd'
                        this_basin = 'A'

                    # Get CVs for this shooting point   # todo: a bit sloppy... can I clean this up?
                    try:
                        if not os.path.exists(thread.history.init_coords[step_index][0]):
                            warnings.warn('attempted to resample ' + thread.history.init_coords[step_index][0] + ' but no such '
                                          'file exists in the working directory\nSkipping and continuing', RuntimeWarning)
                            thread.cvs_for_later.append([])
                            continue        # skip to next step_index
                    except IndexError:  # getting cv's failed (maybe corrupt coordinate file) so consider this step failed
                        thread.cvs_for_later.append([])
                        continue        # skip to next step_index
                    try:
                        this_cvs = get_cvs(thread.history.init_coords[step_index][0], settings)
                    except IndexError:  # getting cv's failed (maybe corrupt coordinate file) so consider this step failed
                        thread.cvs_for_later.append([])
                        continue        # skip to next step_index

                    # Write CVs to as_raw_resample.out
                    f1.write(this_basin + ' <- ' + this_cvs + '\n')
                    f1.flush()
                    if settings.information_error_checking:
                        f2.write(str(thread.history.timestamps[step_index]) + ' ' + this_basin + ' <- ' + this_cvs + '\n')
                        f2.flush()

                    # Append this_cvs to running list for evaluating decorrelation time
                    thread.this_cvs_list.append([[float(item) for item in this_cvs.split(' ')], thread.history.timestamps[step_index]])
                    thread.cvs_for_later.append([float(item) for item in this_cvs.split(' ')])
                else:
                    thread.cvs_for_later.append([])

        # Close files just to be sure
        f1.close()
        if settings.information_error_checking:
            f2.close()

        if settings.information_error_checking:   # sort timestamped output file
            shutil.copy(settings.working_directory + '/as_raw_timestamped.out', settings.working_directory + '/as_raw_timestamped_copy.out')
            open(settings.working_directory + '/as_raw_timestamped.out', 'w').close()
            with open(settings.working_directory + '/as_raw_timestamped_copy.out', 'r') as f:
                for line in sorted(f):
                    open(settings.working_directory + '/as_raw_timestamped.out', 'a').write(line)
                open(settings.working_directory + '/as_raw_timestamped.out', 'a').close()
            os.remove(settings.working_directory + '/as_raw_timestamped_copy.out')

        # Construct list of data lengths to perform decorrelation for
        if settings.information_error_checking:
            if not partial:
                lengths = [leng for leng in range(settings.information_error_freq, len(open(settings.working_directory + '/as_raw_timestamped.out', 'r').readlines()) + 1, settings.information_error_freq)]
            else:   # if partial
                lengths = [leng for leng in range(settings.information_error_freq, len(open(settings.working_directory + '/as_raw_timestamped.out', 'r').readlines()) + 1, settings.information_error_freq) if not leng in [int(line.split(' ')[0]) for line in open(settings.working_directory + '/info_err.out', 'r').readlines()]]
            pattern = re.compile('[0-9]+')  # pattern for reading out timestamp from string
        else:
            lengths = [len(open(settings.working_directory + '/as_raw_resample.out', 'r').readlines())]
            pattern = None

        # Assess decorrelation and write as_decorr.out
        for length in lengths:
            if settings.information_error_checking:
                suffix = '_' + str(length)     # only use-case with multiple lengths, so this keeps them from stepping on one another's toes
                cutoff_timestamp = int(pattern.findall(open(settings.working_directory + '/as_raw_timestamped.out', 'r').readlines()[length - 1])[0])
            else:
                cutoff_timestamp = math.inf
                suffix = ''
            open(settings.working_directory + '/as_decorr' + suffix + '.out', 'w').close()
            f3 = open(settings.working_directory + '/as_decorr' + suffix + '.out', 'a')
            for thread in allthreads:
                if thread.this_cvs_list:       # if there were any 'fwd' or 'bwd' results in this thread
                    mapped = list(map(list, zip(*[item[0] for item in thread.this_cvs_list if item[1] <= cutoff_timestamp])))   # list of lists of values of each CV

                    slowest_lag = -1    # initialize running tally of slowest autocorrelation time among CVs in this thread
                    if settings.include_qdot:
                        ndims = len(thread.this_cvs_list[0]) / 2   # number of non-rate-of-change CVs
                        if not ndims % 1 == 0:
                            raise ValueError('include_qdot = True, but an odd number of dimensions were found in the '
                                             'threads in restart.pkl, so they can\'t contain inertial terms.')
                        ndims = int(ndims)
                    else:
                        ndims = len(thread.this_cvs_list[0])

                    for dim_index in range(ndims):
                        slowest_lag = -1
                        if mapped:
                            this_cv = mapped[dim_index]
                        if len(this_cv) > 1:
                            this_autocorr = stattools.acf(this_cv, nlags=len(this_cv) - 1, fft=True)
                            for lag in range(len(this_cv) - 1):
                                corr = this_autocorr[lag]
                                if abs(corr) <= 1.96 / numpy.sqrt(len(this_cv)):
                                    slowest_lag = lag + 1
                                    break

                    if slowest_lag > 0:     # only proceed to writing decorrelated output file if a slowest_lag was found
                        # Write the same way as to as_raw_resample.out above, but starting the range at slowest_lag
                        for step_index in range(slowest_lag, len(thread.history.prod_results)):
                            if thread.history.prod_results[step_index][0] in ['fwd', 'bwd'] and thread.history.timestamps[step_index] <= cutoff_timestamp:
                                if thread.history.prod_results[step_index][0] == 'fwd':
                                    this_basin = 'B'
                                else:  # 'bwd'
                                    this_basin = 'A'

                                # Get CVs for this shooting point and write them to the decorrelated output file
                                if thread.cvs_for_later[step_index]:
                                    this_cvs = thread.cvs_for_later[step_index]    # retrieve CVs from last evaluation
                                    f3.write(this_basin + ' <- ' + ' '.join([str(item) for item in this_cvs]) + '\n')

            f3.close()

        # Move resample raw output file to take its place as the only raw output file
        shutil.move(settings.working_directory + '/as_raw_resample.out', settings.working_directory + '/as_raw.out')
    else:
        if not full_cvs:
            raise RuntimeError('resample called with only_full_cvs = True but full_cvs = False, so this does nothing. '
                               'Change one of them!')

    # Implement full_cvs
    if full_cvs:
        open(settings.working_directory + '/as_full_cvs.out', 'w').close()
        temp_settings = copy.deepcopy(settings)
        temp_settings.include_qdot = False  # never want to include_qdot in this upcoming call to get_cvs
        try:
            affinity = len(os.sched_getaffinity(0))
        except AttributeError:  # os.sched_getaffinity raises AttributeError on non-UNIX systems.
            affinity = 1
        if affinity == 1:
            with open(settings.working_directory + '/as_full_cvs.out', 'a') as f:
                for thread in allthreads:
                    for step_index in range(min([len(thread.history.prod_results), len(thread.history.prod_trajs)])):    # just in case one got an extra write in over the other
                        if thread.history.prod_results[step_index] in [['fwd', 'bwd'], ['bwd', 'fwd']]:     # if step accepted
                            for job_index in range(2):
                                if os.path.exists(thread.history.prod_trajs[step_index][job_index]):
                                    f.write(get_cvs(thread.history.prod_trajs[step_index][job_index], temp_settings, False, 'all') + '\n')
        else:   # affinity > 1
            # Map partial_full_cvs calls to available processes
            with Pool(affinity) as p:
                p.starmap(partial_full_cvs, zip(allthreads, ['partial_full_cvs_' + str(thread_index) + '.out' for thread_index in range(len(allthreads))], itertools.repeat(temp_settings)))
            # Finally, combine the partial files into the full file
            with open(settings.working_directory + '/as_full_cvs.out', 'w') as outfile:
                for fname in ['partial_full_cvs_' + str(thread_index) + '.out' for thread_index in range(len(allthreads))]:
                    with open(fname) as infile:
                        for line in infile:
                            if line:    # skip blank lines
                                outfile.write(line)
                    os.remove(fname)

def partial_full_cvs(thread, filename, settings):
    """
    Write the full CVs list to the file given by filename for each accepted move in each thread in threads.

    A helper function for parallelizing get_cvs (has to be defined at top-level for multiprocessing)

    Parameters
    ----------
    thread : Thread
        Thread object to evaluate
    filename : str
        Name of the file to write CVs to
    settings : argparse.Namespace
        Settings namespace object

    Returns
    -------
    None

    """
    open(settings.working_directory + '/' + filename, 'w').close()
    with open(settings.working_directory + '/' + filename, 'a') as f:
        for step_index in range(min([len(thread.history.prod_results), len(thread.history.prod_trajs)])):  # just in case one got an extra write in over the other
            if thread.history.prod_results[step_index] in [['fwd', 'bwd'], ['bwd', 'fwd']]:  # if step accepted
                for job_index in range(2):
                    if os.path.exists(thread.history.prod_trajs[step_index][job_index]):
                        f.write(get_cvs(thread.history.prod_trajs[step_index][job_index], settings, False, 'all') + '\n')

def interpret_cv(cv_index, settings):
    """
    Read the cv_index'th CV from settings.cvs, identify its type (distance, angle, dihedral, or differance-of-distances)
    and the atom indices the define it (one-indexed) and return these.

    This function is designed for use in the umbrella_sampling jobtype only. For this reason, it only supports the
    aforementioned CV types. If none of these types appears to fit, this function raises a RuntimeError.

    Parameters
    ----------
    cv_index : int
        The index for the CV to use; e.g., 6 corresponds to CV6. Must be in the range [1, len(settings.cvs)].
    settings : argparse.Namespace
        Settings namespace object

    Returns
    -------
    atoms : list
        A list of 1-indexed atom indices as strings that define the given CV
    optype : str
        A string (either 'distance', 'angle', 'dihedral', or 'diffdistance') corresponding to the type for this CV.
    nat : int
        The number of atoms constituting the given CV

    """
    if not 1 <= cv_index <= len(settings.cvs):
        raise RuntimeError('called interpret_cv with an index outside the range [1, len(settings.cvs)].\n'
                           'len(settings.cvs) = ' + str(len(settings.cvs)) + ' and cv_index = ' + str(cv_index))

    this_cv = settings.cvs[cv_index - 1]  # -1 because CVs are 1-indexed
    if 'pytraj.dihedral' in this_cv or 'mdtraj.compute_dihedrals' in this_cv:
        optype = 'dihedral'
        nat = 4
    elif 'pytraj.angle' in this_cv or 'mdtraj.compute_angles' in this_cv:
        optype = 'angle'
        nat = 3
    elif 'pytraj.distance' in this_cv or 'mdtraj.compute_distances' in this_cv:
        if '-' in this_cv and (this_cv.count('pytraj.distance') == 2 or this_cv.count('mdtraj.compute_distances') == 2 or
                                   ('mdtraj.distance' in this_cv and 'pytraj.distance' in this_cv)):
            optype = 'diffdistance'
            nat = 4
        else:
            optype = 'distance'
            nat = 2
    else:
        raise RuntimeError('unable to discern CV type for CV' + str(int(cv_index)) + '\nOnly '
                           'distances, angles, dihedrals, and differences of distances (all defined '
                           'using either pytraj or mdtraj distance, angle, and/or dihedral functions) '
                           'are supported in umbrella sampling. The offending CV is defined as: ' +
                           this_cv)

    # Get atom indices as a string, then convert to list
    atoms = ''
    if not optype == 'diffdistance':
        count = 0
        for match in re.finditer('[\[\\\']([@0-9]+[,\ ]){' + str(nat - 1) + '}[@0-9]+[\]\\\']',
                                 this_cv.replace(', ', ',')):
            atoms += this_cv.replace(', ', ',')[match.start():match.end()]  # should be only one match
            count += 1
        if not count == 1:
            raise RuntimeError('failed to identify atoms constituting CV definition: ' + this_cv +
                               '\nInterpreted as a ' + optype + ' but found ' + str(count) +
                               ' blocks of atom indices with length ' + str(nat) + '(should be one). Is'
                                                                                   ' this CV formatted in an unusual way?')
        if not atoms:
            raise RuntimeError('unable to identify atoms constituting CV definition: ' + this_cv +
                               '\nIs it formatted in an unusual way?')
        atoms = atoms.replace('[', '').replace(']', ',').replace('\'',
                                                                 '')  # included delimeters for safety, but don't want them
        if '@' in atoms:
            atoms = [item.replace('@', '') for item in atoms.split(' @')]  # pytraj style atom indices
        else:
            atoms = atoms.split(',')  # mdtraj style atom indices
            atoms = [str(int(item) + 1) for item in atoms if not item == '']  # fix zero-indexing in mdtraj
    else:
        count = 0
        for match in re.finditer('[\[\\\']([@0-9]+[,\ ]){1}[@0-9]+[\]\\\']', this_cv.replace(', ', ',')):
            atoms += this_cv.replace(', ', ',')[match.start():match.end()]  # should be two matches
            count += 1
        if not count == 2:
            raise RuntimeError('failed to identify atoms constituting CV definition: ' + this_cv +
                               '\nInterpreted as a difference of distances but found ' + str(count) +
                               ' blocks of atom indices with length 2 (should be two). Is this CV '
                               'formatted in an unusual way?')
        if not atoms:
            raise RuntimeError('unable to identify atoms constituting CV definition: ' + this_cv +
                               '\nIs it formatted in an unusual way?')
        atoms = atoms.replace('[', '').replace(']', ',').replace('\'', '')  # included delimeters for safety, but don't want them
        if '@' in atoms:
            atoms = [item.replace('@', '') for item in atoms.replace('\'', ' ').replace('@', ' @').split()]
        else:
            atoms = atoms.split(',')
            atoms = [str(int(item) + 1) for item in atoms if not item == '']  # fix zero-indexing in mdtraj

    while '' in atoms:
        atoms.remove('')  # remove empty list elements if present

    return atoms, optype, nat


def plot_cvs_vs_rc(full_cvs_file, rc_definition, as_output_file):
    """
    Produce a plot of the values of each CV in the RC as a function of RC value from either aimless shooting or umbrella
    sampling data.

    The plot is made in matplotlib but also printed as text to an output file.

    Parameters
    ----------
    full_cvs_file : str
        Path to full_cvs output file (produced by utilities.resample or resampling.resample_umbrella_sampling)
    rc_definition : str
        String representing the definition of the reduced reaction coordinate
    as_output_file : str
        Path to the as_output_file file passed to lmax.py when generating the reaction coordinate

    Returns
    -------
    None

    """
    # Identify relevant CVs
    pattern = re.compile('CV[0-9]+')
    cv_names = pattern.findall(rc_definition)    # get list of CV names, e.g., ['CV6', 'CV35', ...]
    cvs = [int(cv[2:]) - 1 for cv in cv_names]   # get zero-indexed CV indices

    # Get values of relevant CVs from each line in full_cvs_file
    values = [[] for _ in range(len(cvs))]
    for line in open(full_cvs_file, 'r').readlines():
        ii = -1
        for cv in cvs:
            ii += 1
            values[ii].append(float(line.split()[cv]))     # [[obs1cv1, obs2cv1], [obs1cv2, obs2cv2]]

    # Get data from input file, and determine minimum and maximum values for each CV, get reduced CV values
    input_data = [[float(item) for item in
                   line.replace('A <- ', '').replace('B <- ', '').replace(' \n', '').replace('\n', '').split(' ')]
                  for line in open(as_output_file, 'r').readlines()]       # [[obs1cv1, obs1cv2], [obs2cv1, obs2cv2]]
    mapped = list(map(list, zip(*input_data)))                  # [[obs1cv1, obs2cv1], [obs1cv2, obs2cv2]]
    minmax = [[numpy.min(item) for item in mapped], [numpy.max(item) for item in mapped]]    # [[mincv1, mincv2], [maxcv1, maxcv2]]
    n_cvs = len(minmax[0])  # number of CVs total in as_output_file, for use later
    minmax = [[minmax[0][ii] for ii in cvs], [minmax[1][ii] for ii in cvs]]     # cut down to just the relevant values
    reduced_values = [[(values[ii][jj] - minmax[0][ii]) / (minmax[1][ii] - minmax[0][ii]) for ii in range(len(values))] for jj in range(len(values[0]))]
    reduced_values = list(map(list, zip(*reduced_values)))  # transpose to match shape of values

    rcs = []     # list of RC values for each sample
    for jj in range(len(reduced_values[0])):    # iterate over samples
        cvs_list = [0 for ii in range(n_cvs)]    # evaluate_rc expects an entry for every CV, not just those in the RC
        for kk in range(len(cvs)):
            cvs_list[cvs[kk]] = reduced_values[kk][jj]
        rcs.append(evaluate_rc(rc_definition, cvs_list))

    # Bin each RC value and evaluate average and stdev of each CV value in each bin
    bins = numpy.linspace(min(rcs), max(rcs), 20) # todo: should the number of bins be a setting?

    assert len(values[0]) == len(rcs)   # sanity check

    means = [[numpy.mean([values[cv][jj] for jj in range(len(rcs)) if bins[bin_ii] <= rcs[jj] <= bins[bin_ii + 1]]) for
              bin_ii in range(len(bins) - 1)] for cv in range(len(values))]
    stds = [[numpy.std([values[cv][jj] for jj in range(len(rcs)) if bins[bin_ii] <= rcs[jj] <= bins[bin_ii + 1]]) for
              bin_ii in range(len(bins) - 1)] for cv in range(len(values))]
    reduced_means = [[numpy.mean([reduced_values[cv][jj] for jj in range(len(rcs)) if bins[bin_ii] <= rcs[jj] <= bins[bin_ii + 1]]) for
              bin_ii in range(len(bins) - 1)] for cv in range(len(values))]
    reduced_stds = [[numpy.std([reduced_values[cv][jj] for jj in range(len(rcs)) if bins[bin_ii] <= rcs[jj] <= bins[bin_ii + 1]]) for
              bin_ii in range(len(bins) - 1)] for cv in range(len(values))]

    # Write results to output file and plot in both reduced and non-reduced forms
    with open('cvs_vs_rc.out', 'w') as f:
        f.write('RC value')
        for name in cv_names:
            f.write('; ' + name + ' mean')
            f.write('; ' + name + ' st.dev.')
        for name in cv_names:
            f.write('; reduced ' + name + ' mean')
            f.write('; reduced ' + name + ' st.dev.')
        f.write('\n')
        for ii in range(len(bins) - 1):
            f.write('%.3f' % numpy.mean([bins[ii], bins[ii + 1]]))
            for jj in range(len(means)):
                f.write('\t' + '%.3f' % float(means[jj][ii]))
                f.write('\t' + '%.3f' % float(stds[jj][ii]))
            for jj in range(len(means)):
                f.write('\t' + '%.3f' % float(reduced_means[jj][ii]))
                f.write('\t' + '%.3f' % float(reduced_stds[jj][ii]))
            f.write('\n')

    # todo: optionally, finish implementing with matplotlib
    # rc_means = [numpy.mean([bins[ii], bins[ii + 1]]) for ii in range(len(bins) - 1)]
    # for ii in range(len(bins) - 1):
    #     plt.plot(rc_means, means)


"""
configure.py
Takes user input file and returns settings namespace object
"""

import argparse
import pytraj       # to support pytraj calls in input file
import mdtraj       # to support mdtraj calls in the input file
import numpy        # to support numpy  calls in input file
import numpy as np  # to support numpy  calls even if called as np
import sys
import os
import shutil
import pickle
import django.template as template
import typing
import pydantic
from atesa import auto_cvs
from django.conf import settings as django_settings

def configure(input_file, user_working_directory=''):
    """
    Configure the settings namespace based on the config file.

    Parameters
    ----------
    input_file : str
        Name of the configuration file to read
    user_working_directory : str
        User override for working directory (overrides value in input_file), ignored if set to ''

    Returns
    -------
    settings : argparse.Namespace
        Settings namespace object

    """

    class Settings(pydantic.BaseModel):
        # This class initializes the settings object with type hints. After being built, it gets exported as an
        # argparse.Namelist object, just for convenience.

        # Core settings required for all jobs
        job_type: str
        batch_system: str
        restart: bool
        md_engine: str = 'amber'
        task_manager: str = 'simple'
        topology: str
        working_directory: str
        overwrite: bool

        # Batch template settings
        init_nodes: int = 1
        init_ppn: int = 1
        init_mem: str = '4000mb'
        init_walltime: str = '00:30:00'
        init_solver: str = 'sander'
        init_extra: str = ''
        prod_nodes: int = 1
        prod_ppn: int = 8
        prod_mem: str = '4000mb'
        prod_walltime: str = '02:00:00'
        prod_solver: str = 'sander'
        prod_extra: str = ''

        # File path settings (required for all jobs, but do have sensible defaults)
        path_to_input_files: str = os.path.dirname(os.path.realpath(__file__)) + '/data/input_files'
        path_to_templates: str = os.path.dirname(os.path.realpath(__file__)) + '/data/templates'

        # CV options; required only for aimless shooting, equilibrium path sampling, and committor analysis
        cvs: typing.List[str] = ['']
        auto_cvs_radius: float = 5
        auto_cvs_exclude_water: bool = True
        auto_cvs_exclude_hydrogen: bool = True
        auto_cvs_type: str = 'mdtraj'   # pytraj or mdtraj used in auto_cvs
        include_qdot: bool = True
        as_settings_file: str = ''

        # Required only for aimless shooting and equilibrium path sampling
        initial_coordinates: typing.List[str] = ['']

        # Required only for aimless shooting and committor analysis
        commit_fwd: typing.Tuple[typing.List[int], typing.List[int], typing.List[float], typing.List[str]] = ([-1], [-1], [-1], ['unset'])
        commit_bwd: typing.Tuple[typing.List[int], typing.List[int], typing.List[float], typing.List[str]] = ([-1], [-1], [-1], ['unset'])

        # Required only for committor analysis, umbrella sampling, and equilibrium path sampling
        rc_definition: str = ''
        as_out_file: str = 'as_raw.out'
        rc_reduced_cvs: bool = True

        # Required only for aimless shooting
        min_dt: int = 1
        max_dt: int = 10
        always_new: bool = True
        full_cvs: bool = False
        only_full_cvs: bool = False
        degeneracy: int = 1
        cleanup: bool = True   # todo: implement asking the user what they want this to be at install time; OR make it a required option (no default); OR add a warning when it's set to False about what that entails for resampling
        # todo: make a script that implements cleanup and that can be called separately so users can cleanup manually whenever desired
        information_error_checking: bool = True
        information_error_threshold: float = 0.1
        information_error_freq: int = 250
        information_error_override: bool = False
        information_error_max_dims: int = 6
        information_error_lmax_string = '--two_line_test'
        max_moves: int = -1     # also used by find_ts
        max_consecutive_fails: int = 10
        sigfigs: int = 3

        # Required only for committor analysis
        committor_analysis_n: int = 10
        committor_analysis_use_rc_out: bool = False
        path_to_rc_out: str = sys.path[0] + '/atesa/tests/test_data/rc.out'
        rc_threshold: float = 0.05
        transmission_coefficient: int = 1

        # Required only for equilibrium path sampling
        eps_rc_min: float = -12
        eps_rc_max: float = 12
        eps_rc_step: float = 1
        eps_rc_overlap: float = 0.1
        eps_n_steps: int = 6
        eps_out_freq: int = 1
        eps_dynamic_seed: typing.Union[int, list] = 20  # int or list (int -> [int for window in eps_windows]; 0 or empty list turns off)
        samples_per_window: int = -1

        # Required only for umbrella sampling
        us_implementation: str = 'plumed'
        us_rc_min: float = -12
        us_rc_max: float = 12
        us_rc_step: float = 0.25
        us_restraint: float = 50
        us_degeneracy: int = 5
        us_auto_coords_directory: str = ''
        us_pathway_restraints_file: str = ''

        # Resampling
        resample: bool = False

        # Required only if restart = True
        restart_terminated_threads: bool = False

        # Not expected to be set by user
        DEBUG: bool = False     # True causes some functions to return dummy values for testing purposes
        pid: int = -1           # process ID of information_error call used in aimless shooting
        information_error_overdue: bool = False    # used for handling information_error calls cleanly
        dont_dump: bool = False     # when True, prevents dumping settings to settings.pkl
        suppress_us_warning = False     # used to prevent repeatedly issuing the same warning during some US runs
        previous_cvs = ''       # for checking whether a full resample is required when restarting based on a change in cvs
        previous_information_error_max_dims = -1    # for checking whether a full resample is required when restarting based on a change in information_error_max_dims
        previous_information_error_lmax_string = '--two_line_test'  # for checking whether a full resample is required when restarting based on a change in information_error_lmax_string

    # Import config file line-by-line using exec()
    try:
        lines = open(input_file, 'r').readlines()
    except FileNotFoundError:
        try:
            lines = open('atesa/' + input_file, 'r').readlines()     # for testing
        except:
            lines = open(input_file, 'r').readlines()   # to reproduce original error
    line_index = 0
    for line in lines:      # each line in the input file is just python code setting a variable;
        line_index += 1
        try:
            exec(line)      # this means that comments are supported using '#' and whitespace is ignored.
        except Exception as e:
            raise ValueError('error raised while reading line ' + str(int(line_index)) + ' of configuration file '
                             + input_file + ': ' + str(e))

    # Define settings namespace to store all these variables
    config_dict = {}
    config_dict.update(locals())
    settings = argparse.Namespace()
    settings.__dict__.update(Settings(**config_dict))

    # Override working directory if provided with user_working_directory
    if user_working_directory:
        settings.working_directory = user_working_directory

    # Format directories properly (no trailing '/')
    if settings.working_directory[-1] == '/':
        settings.working_directory = settings.working_directory[:-1]
    if settings.path_to_input_files[-1] == '/':
        settings.path_to_input_files = settings.path_to_input_files[:-1]
    if settings.path_to_templates[-1] == '/':
        settings.path_to_templates = settings.path_to_templates[:-1]
    if settings.path_to_rc_out[-1] == '/':
        settings.path_to_rc_out = settings.path_to_rc_out[:-1]
    try:
        if settings.us_auto_coords_directory[-1] == '/':
            settings.us_auto_coords_directory = settings.us_auto_coords_directory[:-1]
    except IndexError:  # no directory given, not a problem
        pass

    # Certain functions need working_directory to be an absolute path, so we'll prepend the current path if necessary
    if not settings.working_directory[0] == '/':
        settings.working_directory = os.getcwd() + '/' + settings.working_directory

    # Set Django template environment
    if os.path.exists(settings.path_to_templates):
        settings.env = template.Engine(dirs=[settings.path_to_templates])
        if not django_settings.configured:  # need to configure just once
            django_settings.configure()
    else:
        sys.exit('Error: could not locate templates folder: ' + settings.path_to_templates)

    # Initialize EPS settings based on contents of config file
    if settings.job_type == 'equilibrium_path_sampling':
        eps_lower_boundaries = numpy.arange(settings.eps_rc_min, settings.eps_rc_max, settings.eps_rc_step)
        settings.eps_bounds = []
        for lower_bound in eps_lower_boundaries:
            settings.eps_bounds.append([lower_bound - settings.eps_rc_overlap, lower_bound + settings.eps_rc_step + settings.eps_rc_overlap])

        if settings.eps_dynamic_seed:
            if isinstance(settings.eps_dynamic_seed, int):
                settings.eps_dynamic_seed = [settings.eps_dynamic_seed for null in settings.eps_bounds]
            settings.eps_empty_windows = settings.eps_dynamic_seed

    # Obtain cvs and commitment definitions from an existing settings file or call auto_cvs as needed
    if settings.as_settings_file:
        try:
            as_settings = pickle.load(open(settings.as_settings_file, 'rb'))
        except FileNotFoundError:
            raise FileNotFoundError('could not find as_settings_file: ' + settings.as_settings_file)
        except pickle.UnpicklingError:
            raise RuntimeError('provided as_settings_file: ' + settings.as_settings_file + ' could not be loaded as a '
                               'pickle file. Are you sure this is the correct file?')
        settings.cvs = as_settings.cvs
        settings.commit_fwd = as_settings.commit_fwd
        settings.commit_bwd = as_settings.commit_bwd
        settings.include_qdot = as_settings.include_qdot
    elif settings.auto_cvs_radius > 0 and not settings.job_type == 'find_ts':
        settings.cvs = auto_cvs.main(settings=settings)

    return settings

if __name__ == "__main__":
    configure('','')



#!/usr/bin/env python3

"""
Standalone script for converting umbrella sampling output files into free energy profiles via pymbar
"""

import os
import sys
import numpy as np, numpy
import scipy
import pymbar
import glob
import re
import copy
import argparse
import warnings
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.colors as colors


def update_progress(progress, message='Progress'):
    """
    Print a dynamic progress bar to stdout.

    Credit to Brian Khuu from stackoverflow, https://stackoverflow.com/questions/3160699/python-progress-bar

    Parameters
    ----------
    progress : float
        A number between 0 and 1 indicating the fractional completeness of the bar. A value under 0 represents a 'halt'.
        A value at 1 or bigger represents 100%.
    message : str
        The string to precede the progress bar (so as to indicate what is progressing)

    Returns
    -------
    None

    """
    barLength = 10  # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done!\r\n"
    block = int(round(barLength * progress))
    text = "\r" + message + ": [{0}] {1}% {2}".format(
        "#" * block + "-" * (barLength - block), round(progress * 100, 2), status)
    sys.stdout.write(text)
    sys.stdout.flush()


def main(**kwargs):
    """
    Main function for mbar.py. Collect all files in the present directory whose names begin with "rcwin_" and end with
    "us.dat", interpret them as umbrella sampling data centered on the RC value indicated by the first
    underline-separated float in the name, using the pymbar package and user-provided settings.

    Parameters
    ----------
    kwargs : dict
        Dictionary object containing arguments passed in from command line in "if __name__ == '__main__'" section.

    Returns
    -------
    None

    """

    # Format input directory string to ensure it has a trailing '/'
    input_path = kwargs['i'][0]
    if not input_path[-1] == '/':
        input_path += '/'

    # Compile and sort data files
    data_files = [name.replace(input_path, '') for name in glob.glob(input_path + 'rcwin_*_us.dat') if (len(open(name, 'r').readlines()) - 1) > kwargs['min_data'][0]]
    pattern = re.compile('[-0-9.]+')
    data_files = sorted(data_files, key=lambda x: float(pattern.findall(x)[0]))
    
    os.chdir(input_path)

    if len(data_files) == 0:
        raise RuntimeError('did not find any data files matching the expected naming convention and with length of at '
                           'least min_data (' + str(kwargs['min_data'][0]) + ')')

    # Set thermodynamic parameters
    R = 1.987e-3                        # gas constant in kcal/mol-K
    beta = 1 / (R * kwargs['t'][0])     # inverse temperature of simulations (in 1/(kcal/mol))
    kconst = kwargs['k'][0]             # spring constant in kcal/mol

    # PMF range
    try:                    # try for user-defined rc_min
        rc_min = float(kwargs['rc_min'][0])
    except ValueError:      # default; use smallest window center
        rc_min = float(pattern.findall(data_files[0])[0])
    try:                    # try for user-defined rc_max
        rc_max = float(kwargs['rc_max'][0])
    except ValueError:      # default; use largest window center
        rc_max = float(pattern.findall(data_files[-1])[0])

    # Refresh data files to reflect new min and max, if necessary
    if not rc_max == float(pattern.findall(data_files[-1])[0]) or not rc_min == float(pattern.findall(data_files[0])[0]):
        data_files = [name.replace(input_path, '') for name in glob.glob(input_path + 'rcwin_*_us.dat') if rc_min <= float(pattern.findall(name.replace(input_path, ''))[0]) <= rc_max]
        data_files = sorted(data_files, key=lambda x: float(pattern.findall(x)[0]))

    # # Combine data from identical states into temp files, implementing decorr and ignore in the process
    def reject_outliers(data, m=4):
        # Helper function to remove outliers (more than m stdevs from median) from datasets
        d = np.abs(data - np.median(data))
        mdev = np.median(d)
        s = d / mdev if mdev else 0.
        return data[s < m]

    count = 0
    if not kwargs['quiet']:
        update_progress(0, 'Collecting data')
    data_files_temp = []
    old_temp_files = glob.glob('mbar_temp_*_0.dat')     # collect leftover temp files, if any, to delete
    for file in old_temp_files:
        os.remove(file)
    for data_file in data_files:    # combine all data in a single window into one temp file for that window
        center = str(pattern.findall(data_file)[0])
        if not os.path.exists('mbar_temp_' + center + '_0.dat'):
            open('mbar_temp_' + center + '_0.dat', 'w').close()
        lines = open(data_file, 'r').readlines()[kwargs['ignore'][0] + 1:]
        lines = lines[0:len(lines):20]  # for debugging, not for production
        data = []
        for line in lines:
            try:
                if (len(line.split()) > 1
                        and not line.split()[1] == '-'
                        and line.split()[1].count('.') <= 1
                        and line.split()[0].count('.') <= 1
                        ):
                    data.append(float(line.split()[1]))
            except ValueError:
                warnings.warn('misformated data file: ' + str(data_file) + ', skipping a line...')
                # count += 1
                # continue
        data = np.asarray(data)
        data = reject_outliers(data)

        if kwargs['decorr']:
            [t0, g, Neff_max] = pymbar.timeseries.detectEquilibration(data)  # compute indices of uncorrelated timeseries
            data_equil = data[t0:]
            indices = pymbar.timeseries.subsampleCorrelatedData(data_equil, g=g)
            A_n = data[indices]
        else:
            A_n = data
        with open('mbar_temp_' + center + '_0.dat', 'a') as f:
            for item in A_n:
                f.write(str(item) + '\n')
        if not 'mbar_temp_' + center + '_0.dat' in data_files_temp:
            data_files_temp.append('mbar_temp_' + center + '_0.dat')
        if not kwargs['quiet']:
            count += 1
            update_progress(count / len(data_files), 'Collecting data')
    data_files = copy.deepcopy(data_files_temp)

    K = len(data_files)             # number of umbrellas
    N_max = max(len(open(data_files[k], 'r').readlines()) for k in range(K))    # maximum number of snapshots/simulation

    rc_kn = numpy.zeros([K,N_max])  # rc_kn[k,n] is the RC for snapshot n from umbrella simulation k

    N_k = []    # number of samples from simulation k
    K_k = []    # spring constants
    rc0_k = []  # window centers

    # pymbar wants data in the form of a np.ndarray u_kn of shape [k, n] where there are k states and n samples
    # n.b. that u_kn has every sample in every row, regardless of the "source" window
    centers = []
    alldata = []
    for k in range(K):
        center = float(pattern.findall(data_files[k])[0])

        if rc_min <= center <= rc_max:
            lines = open(data_files[k], 'r').readlines()
            A_n = np.asarray([float(line) for line in lines])

            if not center in centers:
                centers.append(center)
                alldata.append(list(A_n))
            else:
                center_index = centers.index(center)
                alldata[center_index] += list(A_n)

            N_k.append(len(A_n))        # number of decorrelated samples in this window
            K_k.append(kconst)          # constant spring constant
            rc0_k.append(float(pattern.findall(data_files[k])[0]))  # window center from file name
            for n in range(len(A_n)):
                rc_kn[k,n] = A_n[n]        # list of all decorrelated samples in this window

    # Now that all of the necessary data is loaded into memory, delete the temp files
    for filename in data_files_temp:
        os.remove(filename)

    # Deprecated printing of data for debugging
    # print([rc0_k[k] for k in range(K)])
    # print([numpy.mean([item for item in rc_kn[k,:] if not item == 0]) - rc0_k[k] for k in range(K)])
    # print([scipy.stats.sem([item for item in rc_kn[k,:] if not item == 0]) for k in range(K)])

    if not kwargs['quiet']:
        from textwrap import wrap
        fig, ax = plt.subplots()
        ax.errorbar([rc0_k[k] for k in range(K)], [numpy.mean([item for item in rc_kn[k,:] if not item == 0]) - rc0_k[k] for k in range(K)], yerr=[scipy.stats.tstd([item for item in rc_kn[k,:] if not item == 0]) for k in range(K)], color='#0072BD', lw=1)
        plt.ylabel('Mean Value', weight='bold')
        plt.xlabel('Reaction Coordinate', weight='bold')
        title = ax.set_title('Diagnostic Mean Value Plot. Should probably be smooth.\n'
                'See ATESA\'s troubleshooting documentation\nfor further '
                'information (Umbrella Sampling > Mean Value Plot).\n')
        fig.tight_layout()
        title.set_y(1.05)
        fig.subplots_adjust(top=0.8)
        plt.show()

    open(kwargs['o'][0], 'w').close()   # initialize output file (and overwrite it if it exists)
    with open(kwargs['o'][0], 'a') as f:
        f.write('ATESA mbar.py output file. Skip to the end for the free energy profile if desired.\n')
        f.write('\n~~Mean Value Plot~~\n')
        f.write('Diagnostic Mean Value Plot. Should probably be smooth.\n'
                'See ATESA\'s troubleshooting documentation\nfor further '
                'information (Umbrella Sampling > Mean Value Plot).\n')
        f.write('Reaction Coordinate    Mean Value    Standard Deviation of Mean Value\n')
        for k in range(K):
            f.write(str(rc0_k[k]) + '\t' +
                    '%.3f' % numpy.mean([item for item in rc_kn[k,:] if not item == 0]) + '\t' +
                    '%.3f' % scipy.stats.tstd([item for item in rc_kn[k,:] if not item == 0]) + '\n')

    # Deprecated first-value plot
    # fig, ax = plt.subplots()
    # ax.plot([rc0_k[k] for k in range(K)], [rc_kn[k,0] for k in range(K)], color='#0072BD', lw=1)
    # plt.ylabel('First Value', weight='bold')
    # plt.xlabel('Reaction Coordinate', weight='bold')
    # plt.show()

    # Set number of bins for 1D PMF (equal to 1.5 times the number of unique window centers)
    # This is somewhat arbitrary and can be changed if desired, but there should be no need
    nbins = int(1.5 * len(set(rc0_k)))

    # Deprecated all-in-one histogram
    # fig, ax = plt.subplots()
    # n, bins, patches = ax.hist([item for item in np.reshape([rc_kn[k,:] for k in range(K)],-1) if not item == 0], int(10 * nbins))
    # plt.show()

    with open(kwargs['o'][0], 'a') as f:
        f.write('\n~~Sampling Histogram~~\nEach pair of lines is bin edges followed by bin heights\n')

        fig = plt.figure()
        ax = fig.add_subplot(111)

        for this_data in alldata:
            nextcolor = list(colors.to_rgb(next(ax._get_patches_for_fill.prop_cycler).get('color'))) + [0.75]
            n, bins, null = ax.hist(this_data, list(numpy.arange(min(this_data), max(this_data) + 0.05, 0.05)), color=nextcolor)    # todo: fixed bin width of 0.05 is unwise
            if not kwargs['quiet']:
                fig.canvas.draw()
            f.write(' '.join(['%.3f' % item for item in bins]) + '\n')
            f.write(' '.join([str(int(item)) for item in n]) + '\n\n')

        if not kwargs['quiet']:
            plt.xlabel('Reaction Coordinate', weight='bold')
            plt.ylabel('Number of Samples', weight='bold')
            plt.show()

    N_max = numpy.max(N_k)  # largest amount of data in a single simulation
    u_kn = numpy.zeros([K,N_max], numpy.float64)
    u_kln = numpy.zeros([K,K,N_max], numpy.float64) # u_kln[k,l,n] is the reduced potential energy of snapshot n from umbrella simulation k evaluated at umbrella l

    # Bin data
    delta = (rc_max - rc_min) / float(nbins)
    bin_center_i = numpy.zeros([nbins], numpy.float64)
    for i in range(nbins):
        bin_center_i[i] = rc_min + delta/2 + delta * i
    bin_kn = numpy.zeros([K,N_max], numpy.int32)
    for k in range(K):
        for n in range(N_k[k]):
            bin_kn[k,n] = list(bin_center_i).index(min(bin_center_i, key=lambda x: abs(x - rc_kn[k][n])))
            # bin_kn[k,n] = int((rc_kn[k][n] - rc_min) / delta)   # compute bin assignment

    # print(nbins)
    # print(np.min(bin_kn))
    # print(np.max(bin_kn))
    # print(bin_center_i)
    # sys.exit()

    # Evaluate reduced energies in all umbrellas
    count = 0
    update_progress(0, 'Processing data')
    for k in range(K):
        for n in range(N_max):
            for l in range(K):
                # Compute deltaRC value of each point to this window center
                drc = rc_kn[k,n] - rc0_k[l]

                # Compute energy of point n interpreted as from window k with umbrella restraint l
                u_kln[k,l,n] = u_kn[k,n] + beta * K_k[k] * drc**2

        if not kwargs['quiet']:
            count += 1
            update_progress(count / K, 'Processing data')

    if not kwargs['quiet']:
        verbose = True
    else:
        verbose = False
    mbar = pymbar.mbar.MBAR(u_kln, N_k, verbose=verbose, maximum_iterations=1000)
    (f_i, df_i) = mbar.computePMF(u_kn, bin_kn, nbins)

    # I have no idea why the first and last bins always get ludicrous values, but they can be ignored
    f_i = f_i[1:len(f_i) - 1]
    f_i = np.array([f - min(f_i) for f in f_i])
    df_i = df_i[1:len(df_i) - 1]
    bin_center_i = bin_center_i[1:len(bin_center_i) - 1]
    nbins -= 2

    # Normalize by beta to convert from kT's of energy to kcal/mol of energy
    f_i /= beta
    df_i /= beta

    if not kwargs['quiet']:
        print('\nFinal results. This analysis was based on the pyMBAR package. If you publish work based on this '
              'output, in addition to ATESA please cite:\n')
        print('Shirts MR and Chodera JD. Statistically optimal analysis of samples from multiple equilibrium states. J.'
              ' Chem. Phys. 129:124105 (2008). DOI: 10.1063/1.2978177\n')
        if kwargs['decorr']:
            print('Since this run also included the "decorr" option, you also need to cite:\n')
            print('Chodera JD. A simple method for automated equilibration detection in molecular simulations. J. Chem.'
                  ' Theor. Comput. 12:1799, 2016. DOI: 10.1021/acs.jctc.5b00784')
        print("\nPMF (kcal/mol)")
        print("%8s %8s %8s" % ('bin', 'f', 'df'))
        for i in range(nbins):
            print("%8.2f %8.3f %8.3f" % (bin_center_i[i], f_i[i], df_i[i]))

        fig, ax = plt.subplots()
        ax.plot([bin_center_i[i] for i in range(nbins)], [f_i[i] for i in range(nbins)], color='#0072BD', lw=2)
        plt.fill_between([bin_center_i[i] for i in range(nbins)], np.asarray([f_i[i] for i in range(nbins)]) - np.asarray([df_i[i] for i in range(nbins)]),
                         np.asarray([f_i[i] for i in range(nbins)]) + np.asarray([df_i[i] for i in range(nbins)]),
                         alpha=0.5, facecolor='#0072BD')
        plt.ylabel('Free Energy (kcal/mol)', weight='bold')
        plt.xlabel('Reaction Coordinate', weight='bold')
        plt.show()
    with open(kwargs['o'][0], 'a') as f:
        f.write('\nFinal results. This analysis was based on the pyMBAR package. If you publish work based on this '
                'output, in addition to ATESA please cite:\n')
        f.write('Shirts MR and Chodera JD. Statistically optimal analysis of samples from multiple equilibrium states.'
                ' J. Chem. Phys. 129:124105 (2008). DOI: 10.1063/1.2978177\n')
        if kwargs['decorr']:
            f.write('Since this run also included the "decorr" option, you also need to cite:\n')
            f.write('Chodera JD. A simple method for automated equilibration detection in molecular simulations. J. '
                    'Chem. Theor. Comput. 12:1799, 2016. DOI: 10.1021/acs.jctc.5b00784')
        f.write('\n~~Free Energy Profile~~\nReaction coordinate    Free energy (kcal/mol)    Error (kcal/mol)\n')
        for i in range(nbins):
            f.write('%.3f' % bin_center_i[i] + '    ' + '%.3f' % f_i[i] + '    ' + '%.3f' % df_i[i] + '\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Evaluate free energy profile from the given umbrella sampling data',
                                     epilog='For more detailed documentation, see:\n'
                                            'https://atesa.readthedocs.io/en/latest/auxiliary_scripts.html'
                                            '#mbar-py-energy-profiles-from-us')

    parser.add_argument('-t', metavar='temp', type=int, nargs=1, default=[300],
                        help='temperature in Kelvin to evaluate energy at. Default=300')
    parser.add_argument('-k', metavar='kconst', type=float, nargs=1, default=[50.0],
                        help='restraint weight in each window. The default matches the default value for umbrella '
                             'sampling in ATESA. Default=50.0')
    parser.add_argument('-o', metavar='output', type=str, nargs=1, default=['mbar.out'],
                        help='the name of the output file containing the data for the histogram, mean value, and free '
                             'energy plots. This file will be overwritten if it exists. Default=mbar.out')
    parser.add_argument('-i', metavar='input_path', type=str, nargs=1, default=['./'],
                        help='the path to the directory containing the data files. Both relative and absolute paths are'
                             ' accepted. Default=./')
    parser.add_argument('--min_data', metavar='min_data', type=int, nargs=1, default=[0],
                        help='minimum number of lines in a given data file for it to be eligible for inclusion in mbar.'
                             ' Default=0')
    parser.add_argument('--ignore', metavar='ignore', type=int, nargs=1, default=[1],
                        help='Number of samples from beginning of each data file to ignore in the analysis, as time to '
                             'decorrelate from initial coordinates or allow for equilibration. Default=1')
    parser.add_argument('--decorr', action='store_true', default=False,
                        help='use pymbar.timeseries.detectEquilibration and pymbar.timeseries.subsampleCorrelatedData '
                             'to attempt to automatically use only equilibrated and decorrelated data in the analysis.')    # todo: make decorr the default, change this to --no-decorr?
    parser.add_argument('--rc_min', metavar='rc_min', nargs=1, default=[''],
                        help='lower bound for range of RC values to include in the energy profile. The default setting '
                             'automatically uses the smallest window center value available, so only set this option if'
                             ' the default isn\'t working or if you want less than the full profile.')
    parser.add_argument('--rc_max', metavar='rc_max', nargs=1, default=[''],
                        help='upper bound for range of RC values to include in the energy profile. The default setting '
                             'automatically uses the largest window center value available, so only set this option if'
                             ' the default isn\'t working or if you want less than the full profile.')
    parser.add_argument('--quiet', action='store_true', default=False,
                        help='suppress all plots and progress bars. The only output will be the output file.')

    arguments = vars(parser.parse_args())  # Retrieves arguments as a dictionary object

    main(**arguments)



"""
Unit and regression test for information_error.py.
"""

# Import package, test suite, and other packages as needed
import atesa
import sys
import os
import shutil
import pytest
import pickle
import subprocess
import glob
from atesa import information_error
from atesa.configure import configure

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass

    def test_main(self):
        """Tests information_error.main using sham shooting points in test_data"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        settings.include_qdot = False
        settings.DEBUG = True   # to skip resampling and substitute our own as_decorr.out file
        shutil.copy('../test_data/test_velocities_init.rst7', '../test_temp')
        shutil.copy('../test_data/test_two_init.rst7', '../test_temp')
        shutil.copy('../test_data/as.out', '../test_temp')
        shutil.copy('../test_data/as_big.out', 'as_decorr_6054.out')

        # First test without extant settings.pkl file
        with pytest.raises(FileNotFoundError):
            information_error.main()

        settings.__dict__.pop('env')    # env attribute is not picklable
        pickle.dump(settings, open('settings.pkl', 'wb'), protocol=2)   # main will look for this file to load in settings

        shutil.copy('../test_data/as_big.out', 'as_decorr_6054.out')

        information_error.main()
        assert os.path.exists('info_err.out')
        assert float(open('info_err.out', 'r').readlines()[0].split(' ')[0]) == 6054
        assert float('%.2f' % float(open('info_err.out', 'r').readlines()[0].split(' ')[1])) == pytest.approx(0.16, 1E-2)

    @classmethod
    def teardown_method(self, method):
        "Runs at end of class"
        for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
            os.remove(filename)



"""
Unit and regression test for the atesa_v2 package.
"""

# Import package, test suite, and other packages as needed
import atesa
import pytest
import sys
import pytraj
import os
import glob
import shutil
import pickle
from atesa.configure import configure
from atesa import main

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass
    
    def test_atesa_v2_imported(self):
        """Sample test, will always pass so long as import statement worked"""
        assert "atesa" in sys.modules

    def test_main(self):
        """Tests main.main with DEBUG = True to skip actual job submission/monitoring"""
        settings = configure('../../data/atesa.config')
        settings.initial_coordinates = ['../test_data/test.rst7']
        settings.topology = '../test_data/test.prmtop'
        settings.degeneracy = 2
        settings.overwrite = True
        settings.resample = False
        settings.DEBUG = True
        assert main.main(settings) == 'ATESA run exiting normally'

    def test_configure_broken(self):
        """Tests configure.py with a non-existent file"""
        with pytest.raises(FileNotFoundError):
            settings = configure('../../data/ates.config')

    def test_configure_directory(self):
        """Tests configure.py behavior in correcting an improperly formatted directory"""
        shutil.copy('../../data/atesa.config', 'atesa.config')  # todo: make a dedicated 'test' copy of atesa.config for tests to use.
        config_lines = open('atesa.config', 'r').readlines()
        for line in config_lines:
            if 'working_directory' in line:
                line = 'working_directory = \'/foo/bar/\'\n'
            open('atesa_temp.config', 'a').write(line)
        settings = configure('atesa_temp.config')
        assert settings.working_directory == '/foo/bar'

    def test_init_threads_new(self):
        """Tests successful initialization of new threads"""
        settings = configure('../../data/atesa.config')
        settings.initial_coordinates = ['../test_data/test.rst7', '../test_data/test_two_init.rst7']
        settings.degeneracy = 1
        allthreads = main.init_threads(settings)
        assert len(allthreads) == 2
        assert allthreads[0].history.init_inpcrd == ['test.rst7']
        assert allthreads[1].history.init_inpcrd == ['test_two_init.rst7']
        assert allthreads[0].topology == 'test.prmtop'
        assert allthreads[1].topology == 'test.prmtop'

    def test_init_threads_restart(self):
        """Tests successful initialization of restarted threads"""
        settings = configure('../../data/atesa.config')
        # First, make restart.pkl
        settings.degeneracy = 1
        settings.initial_coordinates = ['../test_data/test.rst7', '../test_data/test_two_init.rst7']
        settings.rc_definition = '1.00 + 2.34*CV0 - 5.67*CV1'
        settings.rc_reduced_cvs = False
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        settings.include_qdot = False
        settings.topology = '../test_data/test.prmtop'
        allthreads = main.init_threads(settings)
        allthreads[0].history.init_coords = [['../test_data/test_velocities_init.rst7', '../test_data/test_velocities_init_bwd.rst7'],
                                             ['../test_data/test_two_init.rst7', '../test_data/test_two_init_bwd.rst7']]
        allthreads[0].history.prod_results = [['fwd', 'bwd'], ['bwd', 'bwd']]
        allthreads[1].history.init_coords = [['../test_data/test_velocities_init.rst7', '../test_data/test_velocities_init_bwd.rst7'],
                                             ['../test_data/test_two_init.rst7', '../test_data/test_two_init_bwd.rst7']]
        allthreads[1].history.prod_results = [['bwd', 'bwd'], ['fwd', 'bwd']]
        allthreads[0].history.timestamps = [1, 3]
        allthreads[1].history.timestamps = [2, 4]
        pickle.dump(allthreads, open('restart.pkl', 'wb'), protocol=2)
        # Then, reset settings, set restart = True, and try again
        settings = configure('../../data/atesa.config')
        settings.restart = True
        settings.degeneracy = 1
        allthreads = main.init_threads(settings)
        assert len(allthreads) == 2
        assert allthreads[0].history.prod_results == [['fwd', 'bwd'], ['bwd', 'bwd']]
        assert allthreads[1].history.prod_results == [['bwd', 'bwd'], ['fwd', 'bwd']]
        assert allthreads[0].topology == 'test.prmtop'
        assert allthreads[1].topology == 'test.prmtop'

    def test_thread_get_frame_amber(self):
        """Tests thread.get_frame method with md_engine = 'amber' and frame = -1"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.md_engine = 'amber'
        settings.degeneracy = 1
        allthreads = main.init_threads(settings)
        allthreads[0].jobids.append('01234')
        allthreads[0].history.prod_trajs.append(['../test_data/test.nc'])
        compare_traj = pytraj.iterload('../test_data/test.rst7', allthreads[0].topology)
        test_frame = allthreads[0].get_frame(allthreads[0].history.prod_trajs[0][0], -1, settings)
        query_traj = pytraj.iterload(test_frame, allthreads[0].topology)
        assert query_traj.n_frames == compare_traj.n_frames
        assert pytraj.center_of_mass(query_traj) == pytest.approx(pytraj.center_of_mass(compare_traj), 1e-3)

    def test_import_cvs(self):
        """Tests importing CVs and commitment definitions from a given settings.pkl file"""
        #shutil.copy('../../data/atesa.config', 'temp.config')
        open('temp.config', 'w').write('job_type = \'committor_analysis\'')
        open('temp.config', 'a').write('\nbatch_system = \'slurm\'')
        open('temp.config', 'a').write('\nrestart = False')
        open('temp.config', 'a').write('\ntopology = \'../test_data/test.prmtop\'')
        open('temp.config', 'a').write('\nworking_directory = \'./\'')
        open('temp.config', 'a').write('\noverwrite = False')
        open('temp.config', 'a').write('\nas_settings_file = \'../test_data/settings.pkl\'')
        settings = configure('temp.config')
        assert settings.cvs == ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        assert settings.commit_fwd == [[1, 2], [3, 4], [1.5, 2.0], ['lt', 'gt']]
        assert settings.commit_bwd == [[1, 2], [3, 4], [2.0, 1.5], ['gt', 'lt']]

    @classmethod
    def teardown_method(self, method):
        "Runs at end of each method"
        for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
            os.remove(filename)



"""
Unit and regression test for mbar.py.
"""

# Import package, test suite, and other packages as needed
import atesa
import sys
import pytest
import os
import shutil
import glob
import argparse
import filecmp
from atesa import mbar

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass

    def test_basic_settings(self):
        """Tests mbar.py with basic (default) settings, except quiet = True"""
        for file in glob.glob('../test_data/rcwin_*_us.dat'):
            shutil.copy(file, './')
        kwargs = {'t': [300], 'k': [50], 'o': ['mbar.out'], 'min_data': [0], 'ignore': [1], 'decorr': False, 'rc_min': [''], 'rc_max': [''], 'quiet': True, 'i': ['./']}
        mbar.main(**kwargs)
        assert filecmp.cmp('../test_data/mbar.out', 'mbar.out')

    def test_no_data(self):
        """Tests mbar.py with no valid files in the working directory"""
        kwargs = {'t': [300], 'k': [50], 'o': ['mbar.out'], 'min_data': [0], 'ignore': [1], 'decorr': False,
                  'rc_min': [''], 'rc_max': [''], 'quiet': True, 'i': ['./']}
        with pytest.raises(RuntimeError):
            mbar.main(**kwargs)

    # @classmethod
    # def teardown_method(self, method):
    #     "Runs at end of class"
    #     for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
    #         os.remove(filename)



#!/usr/bin/env python3

"""
Likelihood maximization script. This program is designed to be entirely separable from ATESA in that it can be called
manually to perform likelihood maximization to user specifications and with arbitrary input files; however, it is
required by ATESA's aimless shooting information error convergence criterion.
"""

import sys
import os
import numpy
import time
import math
import itertools
import argparse
import warnings
import pickle
import numdifftools
import statsmodels
from scipy import optimize
from scipy import stats
from scipy.special import erf
import matplotlib.pyplot as plt

try:
    import gnuplotlib
    gnuplot = True
except FileNotFoundError:   # gnuplot not installed
    gnuplot = False

def update_progress(progress, message='Progress', eta=0, quiet=False):
    """
    Print a dynamic progress bar to stdout.

    Credit to Brian Khuu from stackoverflow, https://stackoverflow.com/questions/3160699/python-progress-bar

    Parameters
    ----------
    progress : float
        A number between 0 and 1 indicating the fractional completeness of the bar. A value under 0 represents a 'halt'.
        A value at 1 or bigger represents 100%.
    message : str
        The string to precede the progress bar (so as to indicate what is progressing)
    eta : int
        Number of seconds to display as estimated completion time (converted into HH:MM:SS)
    quiet : bool
        If True, suppresses output entirely

    Returns
    -------
    None

    """

    if quiet:
        return None

    barLength = 10  # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done!          \r\n"
    block = int(round(barLength * progress))
    if eta:
        # eta is in seconds; convert into HH:MM:SS
        eta_h = str(math.floor(eta/3600))
        eta_m = str(math.floor((eta % 3600) / 60))
        eta_s = str(math.floor((eta % 3600) % 60)) + ' '
        if len(eta_m) == 1:
            eta_m = '0' + eta_m
        if len(eta_s) == 2:
            eta_s = '0' + eta_s
        eta_str = eta_h + ':' + eta_m + ':' + eta_s
        text = "\r" + message + ": [{0}] {1}% {2}".format("#" * block + "-" * (barLength - block), round(progress * 100, 2), status) + " ETA: " + eta_str
    else:
        text = "\r" + message + ": [{0}] {1}% {2}".format("#" * block + "-" * (barLength - block), round(progress * 100, 2), status)
    sys.stdout.write(text)
    sys.stdout.flush()


def objective_function(params, A_data, B_data):
    """
    Evaluate the negative log likelihood function for the given parameters and lists of observations.

    This function evaluates the goodness of fit of the given parameters and data to an error function ansatz, as
    described in Peters, 2012. Chem. Phys. Lett. 554: 248.

    Designed to be called by an optimization routine to obtain the best fitting params.

    Parameters
    ----------
    params : list
        Parameters for the current model to be tested
    A_data : list
        List of observations from aimless shooting that committed to basin "A" (usually the reactants)
    B_data : list
        List of observations from aimless shooting that committed to basin "B" (usually the products)

    Returns
    -------
    negative_log_likelihood : float
        The negative log likelihood of the fit to the ansatz for the given parameters and observations

    """

    def erflike(arg):
        pl = numpy.ones(len(arg))
        ml = numpy.negative(numpy.ones(len(arg)))
        return numpy.where(arg > 5.7, pl, numpy.where(arg < -5.7, ml, erf(arg)))

    if A_data and not B_data:
        qa = params[0] + numpy.inner(params[1:], A_data)
        sum = numpy.sum(numpy.log((1 - erflike(qa)) / 2))
    elif B_data and not A_data:
        qb = params[0] + numpy.inner(params[1:], B_data)
        sum = numpy.sum(numpy.log((1 + erflike(qb)) / 2))
    else:
        qa = params[0] + numpy.inner(params[1:], A_data)
        qb = params[0] + numpy.inner(params[1:], B_data)
        sum = numpy.sum(numpy.log((1 - erflike(qa)) / 2)) + numpy.sum(numpy.log((1 + erflike(qb)) / 2))

    return -1 * sum


def two_line_test_func(results, plots, two_line_threshold=0.5):
    """
    Perform a double linear regression on intersecting subsets of the data in results to determine whether to terminate
    and how many dimensions to return in the RC during two_line_test.

    Can only be called with len(results) >= 5.

    Parameters
    ----------
    results : list
        List of dictionary objects indexed by step of two_line_test, each possessing attribute 'fun' giving the optimization
        score for that step
    plots : bool
        If True, plot lines using gnuplot
    two_line_threshold : float
        Ratio of second slope to first slope (as a fraction) below which the two-line test can pass

    Returns
    -------
    out : int
        Index of selected 'best' RC from two-line test; or, -1 if no best RC could be determined

    """

    if len(results) < 5:
        raise RuntimeError('two_line_test can only be called with at least 5 optimized models')

    best_closest = []  # result for which the intersection is closest to the shared point
    for test_index in range(len(results) - 2):  # - 2 to account for minimum of two points in each line
        first_segment = range(1, 3 + test_index)
        second_segment = range(first_segment[-1], len(results) + 1)
        opt1 = stats.linregress(first_segment, [results[i - 1].fun for i in first_segment])
        opt2 = stats.linregress(second_segment, [results[i - 1].fun for i in second_segment])

        # Now evaluate closest point in results to the intersection of the two lines
        x_intersect = (opt1.intercept - opt2.intercept) / (opt2.slope - opt1.slope)
        y_intersect = (opt1.slope * x_intersect) + opt1.intercept
        x_val = 0       # initialize index for keeping track of x values
        min_diff = -1   # initialize smallest distance between intersection and point
        closest = 0     # initialize index of closest point to intersection
        for result in results:
            y_val = result.fun
            x_val += 1
            y_diff = y_val - y_intersect
            x_diff = x_val - x_intersect
            diff = numpy.sqrt(y_diff**2 + x_diff**2)
            if min_diff < 0:
                min_diff = diff
                closest = [x_val, diff]
            elif diff < min_diff:
                min_diff = diff
                closest = [x_val, diff]

        # if the closest point to the intersection is the shared point of the lines;
        if closest[0] == test_index + 2:
            if not best_closest:                    # for the first time
                best_closest = [closest, opt1, opt2]
            elif closest[1] < best_closest[0][1]:   # update the closest yet
                best_closest = [closest, opt1, opt2]

    if gnuplot and plots:
        if len(results[0].x) + 2 == len(results[1].x):  # if this is True, results include rate-of-change terms
            min_dims = (len(results[0].x) - 1) / 2      # smallest model dimensionality to be plotted (-1 for constant)
        else:   # no rate-of-change terms
            min_dims = len(results[0].x) - 1

        points1 = [[i + min_dims for i in range(len(results))],
                   [best_closest[1].slope * (i + 1) + best_closest[1].intercept for i in range(len(results))]]
        points2 = [[i + min_dims for i in range(len(results))],
                   [best_closest[2].slope * (i + 1) + best_closest[2].intercept for i in range(len(results))]]
        gnuplotlib.plot((numpy.asarray([item + min_dims for item in range(len(results))]),
                        numpy.asarray([result.fun for result in results])),
                        (numpy.asarray(points1[0]), numpy.asarray(points1[1]), {'legend': '1st slope: ' + '%.3f' % best_closest[1].slope}),
                        (numpy.asarray(points2[0]), numpy.asarray(points2[1]), {'legend': '2nd slope: ' + '%.3f' % best_closest[2].slope}),
                        _with='lines', terminal='dumb 80,40', unset='grid')
    if plots:
        print('Two_line_test plot data:')
        print(' Model scores: ' + str(numpy.asarray([result.fun for result in results])))
        print(' First line values: ' + str(points1[1]))
        print(' Second line values: ' + str(points2[1]))

    if not best_closest:    # no pairs of lines whose intersection was closest to their shared point
        print('Two line test: found no suitable model, performing an additional optimization step and retrying')
        return -1

    slope_fract = best_closest[2].slope / best_closest[1].slope
    if slope_fract > two_line_threshold:  # best point does not meet threshold for relative difference in slopes
        print('Two line test: best model has ratio of slopes ' + str(slope_fract) + ', which does not meet threshold ' +
              str(two_line_threshold) + '; performing an additional optimization step and retrying')
        return -1
    else:                   # DOES meet threshold; return the index of the passing result
        return best_closest[0][0] - 1   # - 1 because of different indexing standards


def eval_rc(params, obs):
    # Returns reaction coordinate value for a given set of parameters and an observation
    params = list(params)
    rc = params[0]
    for local_index in range(len(obs)):
        rc += params[local_index + 1] * obs[local_index]
    return rc


def main(**kwargs):
    """
    Main runtime function of lmax.py.

    Assembles lists of models to optimize in the form of lists of CVs, passes them to optimize, interprets results, and
    repeats or terminates in accordance with argument-dependent termination criteria.

    Parameters
    ----------
    kwargs : dict
        Dictionary object containing arguments

    Returns
    -------
    None

    """

    # Ensure existence and validity of input file
    input_file = kwargs['i'][0]
    if not os.path.exists(input_file):
        raise FileNotFoundError('could not find input file: ' + input_file)
    input_file_lines = open(input_file, 'r').readlines()
    open(input_file, 'r').close()
    if False in [char == 'A' or char == 'B' for char in [line[0] for line in input_file_lines]]:
        raise RuntimeError('input file ' + input_file + ' does not have \'A\' or \'B\' as the first character in each '
                           'line. Is this the correct file? Be sure to remove any blank lines.')

    # Bring in other arguments, just for neatness
    dims = kwargs['k'][0]
    fixed = kwargs['f']  # we actually want this one to stay a list
    qdot = kwargs['q'][0]
    running = kwargs['r'][0]
    output_file = kwargs['o'][0]
    two_line_test = kwargs['two_line_test']
    plots = kwargs['plots']
    quiet = kwargs['quiet']
    two_line_threshold = kwargs['two_line_threshold'][0]
    skip = kwargs['s']    # this one also a list
    hist_bins = kwargs['hist_bins'][0]
    prefilter = kwargs['p'][0]

    if not fixed == [None] and running == 0 and not two_line_test and len(fixed) > dims:
        raise RuntimeError('value of k must be less than or equal to number of fixed (-f) dimensions.')

    if not fixed == [None] and not skip == [None]:
        if any([f in skip for f in fixed]) or any([s in fixed for s in skip]):
            raise RuntimeError('the same CV cannot be indicated with both the -s and -f options at the same time.')

    # Ignore arguments as described in documentation
    if running:
        if fixed == [None]:
            fixed = []
        dims = running
    if two_line_test:
        if fixed == [None]:
            fixed = []
        dims = -1
        running = 0

    # Load settings object from .pkl file if present, to check for information error override and max_dims
    information_error_max_dims = -1
    if two_line_test:
        try:
            settings = pickle.load(open('settings.pkl', 'rb'))
            if not quiet:
                print('Loaded settings.pkl...')
            try:
                information_error_override = settings.information_error_override
                if not quiet:
                    print('Setting information_error_override = ' + str(information_error_override))
            except AttributeError:
                information_error_override = False
                if not quiet:
                    print('information_error_override is not set; defaulting to False')
            try:
                information_error_max_dims = settings.information_error_max_dims
                if not quiet:
                    print('Setting maximum number of two_line_test dimensions to: ' + str(int(information_error_max_dims)))
            except AttributeError:
                if not quiet:
                    print('information_error_max_dims is not set; defaulting to no limit')
        except FileNotFoundError:
            pass

    # Get data from input file, and determine minimum and maximum values for each CV, reduce data
    input_data = [[float(item) for item in
                   line.replace('A <- ', '').replace('B <- ', '').replace(' \n', '').replace('\n', '').split(' ')]
                  for line in input_file_lines]     # [[obs1cv1, obs1cv2], [obs2cv1, obs2cv2]]
    A_data = [[float(item) for item in line.replace('A <- ', '').replace(' \n', '').replace('\n', '').split(' ')] for
              line in input_file_lines if line[0] == 'A']
    B_data = [[float(item) for item in line.replace('B <- ', '').replace(' \n', '').replace('\n', '').split(' ')] for
              line in input_file_lines if line[0] == 'B']
    mapped = list(map(list, zip(*input_data)))      # [[obs1cv1, obs2cv1], [obs1cv2, obs2cv2]]
    minmax = [[numpy.min(item) for item in mapped], [numpy.max(item) for item in mapped]]    # [[mincv1, mincv2], [maxcv1, maxcv2]]
    N = len(input_file_lines)   # number of observations
    NA = len(A_data)            # number of observations that committed to A...
    NB = len(B_data)            # ... and to B
    num_cvs = len(minmax[0])    # number of CVs recorded in each observation
    reduced_A = [[(A_data[jj][ii] - minmax[0][ii]) / (minmax[1][ii] - minmax[0][ii]) for ii in range(num_cvs)] for jj in range(NA)]
    reduced_B = [[(B_data[jj][ii] - minmax[0][ii]) / (minmax[1][ii] - minmax[0][ii]) for ii in range(num_cvs)] for jj in range(NB)]

    if qdot == 'present' or qdot == 'ignore':
        if not num_cvs % 2 == 0:
            raise RuntimeError('likelihood maximization was attempted with input file: ' + input_file + ' and '
                               'include_qdot (q) = True, but this input file has an odd number of entries per line. Are'
                               ' you sure it includes rate-of-change data?')
        num_cvs = int(num_cvs / 2)

    # Prepare for and then enter optimization loop
    termination = False     # initialize primary termination criterion flag
    termination_2 = False   # additional termination flag for use with qdot = 'present', to perform final optimization
    reached_maximum = False # indicates whether the maximum number of allowed dimensions has been reached by two_line_test
    two_line_result = -1    # initialize current model dimensionality for two_line_test
    cv_combs = [[]]         # initialize list of CV combinations to iterate through
    results = []  # initialize for two_line_test
    prefiltered = []    # initialize results list for prefiltering
    while not termination and len(cv_combs[0]) <= N:
        # Initialize current best result
        current_best = [argparse.Namespace(), [0], [], []]
        current_best[0].fun = math.inf

        # Assemble list of RCs to optimize
        if prefilter < 1:
            cv_combs = [[item] for item in range(1, num_cvs + 1)]
        elif not fixed == [None] and len(fixed) == dims:
            cv_combs = [fixed]
        elif running or two_line_test:
            cv_combs = [fixed + [new] for new in range(1, num_cvs + 1) if (not new in fixed) and (not new in skip)]
        else:
            cv_combs = [comb for comb in itertools.combinations(range(1, num_cvs + 1), dims) if (fixed == [None] or set(fixed).issubset(comb)) and (skip == [None] or not any([skipped in comb for skipped in skip]))]
        if qdot == 'present' and not termination_2:
            cv_combs_temp = cv_combs
            cv_combs = []
            for comb in cv_combs_temp:
                cv_combs.append([])
                for item in comb:
                    cv_combs[-1].append(item)
                    cv_combs[-1].append(item + num_cvs)

        # Perform optimization
        start_params = [0 for null in range(len(cv_combs[0]) + 1)]  # + 1 for constant term
        count = 0
        count_to = len(cv_combs)
        update_progress(0, 'Optimizing ' + str(count_to) + ' combination(s) of CVs', quiet=quiet)
        speed_data = [0,0]
        for comb in cv_combs:
            t = time.time()
            this_A = []
            this_B = []
            for index in comb:  # produce k-by-len(A_data) matrices (list of lists) for the selected CVs
                try:
                    this_A.append([obs[index - 1] for obs in reduced_A])
                except TypeError:
                    print(comb)
                    print(index)
                    raise RuntimeError('user-defined')
                this_B.append([obs[index - 1] for obs in reduced_B])
            this_A = list(map(list, zip(*this_A)))  # transpose the matrices to get desired format
            this_B = list(map(list, zip(*this_B)))
            this_result = optimize.minimize(objective_function, numpy.asarray(start_params), (this_A, this_B),
                                            method='BFGS', options={"disp": False, "maxiter": 20000 * (len(comb) + 1)}) # try SR1?
            if prefilter < 1:
                prefiltered.append([this_result.fun, comb[0]])  # append only comb[0] to ignore qdot terms
            elif this_result.fun < current_best[0].fun:
                current_best = [this_result, comb, this_A, this_B]
            this_speed = time.time() - t
            speed_data = [(speed_data[1] * speed_data[0] + this_speed) / (speed_data[1] + 1), speed_data[1] + 1]
            count += 1
            eta = (count_to - count) * speed_data[0]
            update_progress(count / count_to, 'Optimizing ' + str(count_to) + ' combination(s) of CVs', eta, quiet=quiet)

        # If prefiltering, set 'skip' and step back out to while loop
        if prefilter < 1:
            prefiltered = sorted(prefiltered, key=lambda x: x[0])   # sort by ascending log likelihood
            skip = []
            skip += [entry[1] for entry in prefiltered[int(len(prefiltered) * prefilter):]]
            for fix in fixed:   # add fixed dimensions back in after prefiltering
                if fix in skip:
                    skip.remove(fix)
            prefilter = 1   # so we don't prefilter again
            continue        # go back to the start of the while loop

        # Update fixed and results parameters as needed
        if two_line_test:
            results.append(current_best)
        if running or two_line_test:
            fixed = current_best[1]
            if qdot == 'present':
                for item in fixed:
                    if item > num_cvs:  # remove qdot terms from fixed
                        fixed.remove(item)

        # Check termination criteria
        if not running and not two_line_test:
            termination = True
        elif running and not two_line_test:
            if int(len(current_best[1])) == running:
                termination = True
        elif two_line_test and not termination_2:
            if len(results) >= 5:   # can only confidently check for convergence with at least 5 points
                two_line_result = two_line_test_func([result[0] for result in results], plots, two_line_threshold)
                if two_line_result >= 0:
                    termination = True
                    current_best = results[two_line_result]
        if two_line_test and len(cv_combs[0]) == information_error_max_dims and not termination_2:
            termination = True
            reached_maximum = True
            current_best = results[-1]
        if termination_2:
            termination = True
        if qdot == 'present' and termination and not termination_2:
            termination = False
            termination_2 = True
            fixed = current_best[1]
            for item in fixed:
                if item > num_cvs:  # remove qdot terms from fixed
                    fixed.remove(item)
            dims = len(fixed)

    if two_line_test and (two_line_result < 0 and not reached_maximum):   # ran out of CVs to append and two_line_test never passed
        err = RuntimeError('The two_line_test termination criterion was never satisfied even after including every '
                           'candidate CV in the model reaction coordinate.\nThis almost certainly indicates that either'
                           ' one or more key CVs are absent from the aimless shooting output file supplied, or that not'
                           ' enough unimportant CVs were included to give context to the important ones. Either way you'
                           ' should add more CVs to the list.\nThis error can by bypassed by running lmax.py in a '
                           'directory containing a settings.pkl file with the line "information_error_override = True" '
                           '(without quotes). If you did supply this setting, then you are seeing this message because '
                           'the settings.pkl file could not be found.')
        try:
            if information_error_override:
                pass
            else:
                raise err
        except NameError:
            raise err

    # Calculate hess and jaco using the model in current_best (current_best[2] and [3] are corresponding this_A and this_B)
    l_objective_function = lambda x: objective_function(x, current_best[2], current_best[3])
    hess = numdifftools.Hessian(l_objective_function)(current_best[0].x)

    # jaco has to be a sum of the jacobian transpose times the jacobian over each individual observation in the data
    if not quiet:
        count = 0
        update_progress(0, 'Calculating mean information error')
    total_len = len(current_best[2]) + len(current_best[3])
    jaco = 0
    for this_A in current_best[2]:
        l_objective_function = lambda x: objective_function(x, [this_A], [])
        this_jaco = numdifftools.Jacobian(l_objective_function)(current_best[0].x)
        jaco += numpy.matmul(numpy.transpose(this_jaco), this_jaco)
        if not quiet:
            count += 1
            update_progress(count/total_len, 'Calculating mean information error')
    for this_B in current_best[3]:
        l_objective_function = lambda x: objective_function(x, [], [this_B])
        this_jaco = numdifftools.Jacobian(l_objective_function)(current_best[0].x)
        jaco += numpy.matmul(numpy.transpose(this_jaco), this_jaco)
        if not quiet:
            count += 1
            update_progress(count/total_len, 'Calculating mean information error')

    V = numpy.matmul(numpy.matmul(numpy.linalg.inv(numpy.negative(hess)), jaco), numpy.linalg.inv(numpy.negative(hess)))  # Godambe Information
    weights = [0] + [1 / (len(V[0]) - 1) for null in range(len(V[0]) - 1)]  # weights for mean excluding constant term
    mean_std = numpy.inner(weights, [numpy.sqrt(item) for item in numpy.diag(V)])   # mean of estimated standard errors

    # Return output in desired format
    rc_string = str('%.3f' % current_best[0].x[0]) + ' + ' + ' + '.join(['%.3f' % current_best[0].x[i+1] + '*CV' +
                        str(current_best[1][i]) for i in range(len(current_best[1]))])
    output_string = 'Likelihood maximization complete!\n' \
                    'The optimized reaction coordinate (with CVs indexed from 1) is: ' + rc_string + '\n' \
                    'The negative log likelihood of this model is: ' + '%.3f' % current_best[0].fun + '\n' \
                    'The mean information error for this model is: ' + '%.3f' % mean_std

    if output_file:
        open(output_file, 'w').write(output_string)
    else:
        print(output_string)

    if plots:
        A_results = []
        for obs in current_best[2]:  # iterate over A observations
            A_results.append(eval_rc(current_best[0].x, obs))
        B_results = []
        for obs in current_best[3]:  # iterate over B observations
            B_results.append(eval_rc(current_best[0].x, obs))
        hist_result = numpy.histogram(A_results + B_results, hist_bins)  # this step just to bin, not the final histogram
        rc_values = []      # initialize results list
        probs = []          # initialize results list
        for bin_index in range(len(hist_result[0])):
            A_count = 0
            B_count = 0
            for result in A_results:
                if hist_result[1][bin_index] <= result < hist_result[1][bin_index + 1]:
                    A_count += 1
            for result in B_results:
                if hist_result[1][bin_index] <= result < hist_result[1][bin_index + 1]:
                    B_count += 1
            if A_count or B_count:  # if there is data in this bin
                count_ratio = B_count / (A_count + B_count)
            else:
                count_ratio = 0
                warnings.warn('One or more histogram bins is empty and is being reported errantly as having a committor'
                              ' probability of zero. This indicates either insufficient data in the input file, too '
                              'many histogram bins (the default is 10), or that a CV included in the final RC takes on '
                              'discrete values instead of continuous ones (which is usually inappropriate for lmax).')
            rc_values.append(numpy.mean([hist_result[1][bin_index + 1], hist_result[1][bin_index]]))
            probs.append(count_ratio)

        fig = plt.figure()             # initialize matplotlib figure
        ax = fig.add_subplot(111)       # add axes to the figure
        plt.ylabel('Probability of Commitment to Forward Basin', weight='bold')
        plt.xlabel('Reaction Coordinate', weight='bold')
        ax.bar(rc_values, probs, width=0.9*(rc_values[1] - rc_values[0]), color='#00274C')
        ax.plot(rc_values, (1 + erf(numpy.array([value for value in rc_values])))/2, color='#FFCB05', linewidth=3)
        ax.legend(['Ideal', 'Observed'])

        print('Committor sigmoid histogram data:')
        print(' RC values: ' + str(rc_values))
        print(' Observed probabilities of commitment to the forward basin: ' + str(probs))
        print(' Ideal committor sigmoid: ' + str(list((1 + erf(numpy.array([value for value in rc_values])))/2)))

        fig.canvas.draw()
        plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Perform LMAX on the given input data')
    parser.add_argument('-i', metavar='input_file', type=str, nargs=1, default=['as_decorr.out'],
                        help='Input filename (output from aimless shooting). Default=as_decorr.out')
    parser.add_argument('-k', metavar='dimensionality', type=int, nargs=1, default=[1],
                        help='Number of CVs to include in RC. Default=1')
    parser.add_argument('-f', metavar='fixed', type=int, nargs='*', default=[None],
                        help='CVs to require inside the RC. Default=none')
    parser.add_argument('-s', metavar='skip', type=int, nargs='*', default=[None],
                        help='CVs to skip (not consider in RC). Default=none')
    parser.add_argument('-q', metavar='include_qdot', type=str, nargs=1, default=['present'],
                        help='Valid options are: "present", "absent", and "ignore" (quotes excluded). If "present" or '
                             '"ignore", the input file is assumed to include rate-of-change ("q") data for each CV '
                             '(formatted as in e.g., "A <- CV0 CV1 q0 q1"); in the former case, q terms will be used to'
                             'select the RC (but will not appear in the final RC), implementing inertial likelihood '
                             'maximization. In the latter, rate of change terms are not used. Finally, if "absent", the'
                             ' q data will be assumed not to be present in the input file at all. Default=present')
    parser.add_argument('-r', metavar='running', type=int, nargs=1, default=[0],
                        help='If > 0, runs from k = 1 to "running" using the previously obtained k - 1 results as the '
                             'argument for f, ignoring the arguments passed for k and f. Default=0')
    parser.add_argument('-p', metavar='prefilter', type=float, nargs=1, default=[1],
                        help='A number between 0 and 1. If < 1, every one-term RC is tested before anything else and '
                             'then only the top \'prefilter\' fraction of CVs are considered moving forward. Default=1')
    parser.add_argument('-o', metavar='output_file', type=str, nargs=1, default=[''],
                        help='Prints output to a new file whose name is given with this argument, instead of directly '
                             'to the terminal. The file will be overwritten if it exists. Default=none')
    parser.add_argument('--quiet', action='store_true',
                        help='If this option is given, progress messages outputted to the terminal are suppressed and ' 
                             'only the final result is written (either to the terminal or the output file.)')
    parser.add_argument('--two_line_test', action='store_true', default=False,
                        help='If this option is given, arguments passed for k, f, and r are ignored, and the RC is '
                             'chosen based on the two-line method (see documentation).')
    parser.add_argument('--plots', action='store_true', default=False,
                        help='If True, plots the final fit between the model and data committor sigmoid. '
                             'If this option is given alongside two_line_test, gnuplot will be used to write plots to '
                             'the terminal during evaluations of the two_line_test termination criterion (if it is '
                             'installed). The sigmoid data is also printed to the terminal or output file.')
    parser.add_argument('--two_line_threshold', metavar='two_line_threshold', type=float, nargs=1, default=[0.5],
                        help='If this option is given alongside two_line_test, sets the maximum ratio of slopes in the'
                             'two-line test. See the documentation for two_line_test for details. Default=0.5')
    parser.add_argument('--hist_bins', metavar='hist_bins', type=int, nargs=1, default=[10],
                        help='If this option is given alongside plots, sets the number of reaction coordinate bins for'
                             'the sigmoid committor histogram. Production of the histogram will fail if any of the '
                             'bins have zero samples in them, which is more likely for larger values of hist_bins. '
                             'Default = 10')

    arguments = vars(parser.parse_args())  # Retrieves arguments as a dictionary object

    # Suppress numpy.log and numdifftools/limits.py warnings that occur frequently during normal operation
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='invalid value encountered in less')
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='invalid value encountered in greater')
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='divide by zero encountered in log')
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='invalid value encountered in double_scalars')
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='invalid value encountered in subtract')
    warnings.filterwarnings('ignore', category=RuntimeWarning, message='divide by zero encountered in double_scalars')

    main(**arguments)



"""
Unit and regression test for auto_cvs.py.
"""

# Import package, test suite, and other packages as needed
import atesa
import sys
import os
import shutil
import pytest
import glob
import filecmp
import argparse
import mdtraj
import pytraj
import numpy
import scipy
from atesa import auto_cvs
from atesa.configure import configure

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass

    def test_main_mdtraj(self):
        """Tests auto_cvs with mdtraj"""
        settings = argparse.Namespace()
        settings.auto_cvs_type = 'mdtraj'
        settings.topology = '../test_data/test.prmtop'
        settings.initial_coordinates = ['../test_data/test.rst7']
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        settings.commit_fwd = ([101, 102, 102], [103, 104, 105], [1.5, 2.0, 1.0], ['lt', 'gt', 'lt'])
        settings.commit_bwd = ([101, 102, 102], [103, 104, 105], [2.0, 1.5, 1.8], ['gt', 'lt', 'gt'])
        settings.working_directory = './'
        settings.auto_cvs_radius = 5
        settings.auto_cvs_exclude_water = True
        settings.auto_cvs_exclude_hydrogen = True

        mtraj = mdtraj.load(settings.initial_coordinates[0], top=settings.topology)  # for testing third assert

        result = auto_cvs.main(settings)
        compare = open('../test_data/cvs_mdtraj.txt', 'r').readlines()[1:]

        assert len(result) == len(compare)
        assert all([item in [line.split('; ')[1].strip('\n') for line in compare] for item in result])
        assert pytest.approx(eval(compare[0].split('; ')[1]), 1.09, 1E-2)

    def test_main_pytraj(self):
        """Tests auto_cvs with pytraj"""
        settings = argparse.Namespace()
        settings.auto_cvs_type = 'pytraj'
        settings.topology = '../test_data/test.prmtop'
        settings.initial_coordinates = ['../test_data/test.rst7']
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        settings.commit_fwd = ([101, 102, 102], [103, 104, 105], [1.5, 2.0, 1.0], ['lt', 'gt', 'lt'])
        settings.commit_bwd = ([101, 102, 102], [103, 104, 105], [2.0, 1.5, 1.8], ['gt', 'lt', 'gt'])
        settings.working_directory = './'
        settings.auto_cvs_radius = 5
        settings.auto_cvs_exclude_water = True
        settings.auto_cvs_exclude_hydrogen = False

        traj = pytraj.load(settings.initial_coordinates[0], settings.topology)  # for testing third assert

        result = auto_cvs.main(settings)
        compare = open('../test_data/cvs_pytraj.txt', 'r').readlines()[1:]

        assert len(result) == len(compare)
        assert all([item in [line.split('; ')[1].strip('\n') for line in compare] for item in result])
        assert pytest.approx(eval(compare[0].split('; ')[1]), 1.09, 1E-2)

    @classmethod
    def teardown_method(self, method):
        "Runs at end of class"
        for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
            os.remove(filename)




"""
Unit and regression test for utilities.py.
"""

# Import package, test suite, and other packages as needed
import atesa
import sys
import pytest
import pytraj
import os
import glob
import filecmp
import shutil
import pickle
from atesa import utilities
from atesa.configure import configure
from atesa import process
from atesa import main

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass
    
    def test_check_commit_fwd(self):
        """Tests check_commit using a dummy coordinate file and commitments defined to get result 'fwd'"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.commit_fwd = [[1, 2], [3, 4], [1.0, 1.7], ['gt', 'lt']]
        settings.commit_bwd = [[1, 2], [3, 4], [0.5, 2.0], ['lt', 'gt']]
        assert utilities.check_commit('../test_data/test.rst7', settings) == 'fwd'

    def test_check_commit_bwd(self):
        """Tests check_commit using a dummy coordinate file and commitments defined to get result 'bwd'"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.commit_bwd = [[1, 2], [3, 4], [1.0, 1.7], ['gt', 'lt']]
        settings.commit_fwd = [[1, 2], [3, 4], [0.5, 2.0], ['lt', 'gt']]
        assert utilities.check_commit('../test_data/test.rst7', settings) == 'bwd'

    def test_check_commit_none(self):
        """Tests check_commit using a dummy coordinate file and commitments defined to get result ''"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.commit_fwd = [[1, 2], [3, 4], [1.0, 1.5], ['gt', 'lt']]
        settings.commit_bwd = [[1, 2], [3, 4], [0.5, 2.0], ['lt', 'gt']]
        assert utilities.check_commit('../test_data/test.rst7', settings) == ''

    def test_check_commit_traj(self):
        """Tests check_commit using a dummy trajectory file and commitments defined to get result 'fwd'"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.commit_fwd = [[1, 2], [3, 4], [1.0, 1.7], ['gt', 'lt']]
        settings.commit_bwd = [[1, 2], [3, 4], [0.5, 2.0], ['lt', 'gt']]
        assert utilities.check_commit('../test_data/test.nc', settings) == 'fwd'

    def test_check_commit_value_errors(self):
        """Tests check_commit using a dummy coordinate file and commitments defined to get ValueErrors"""
        settings = configure('../../data/atesa.config')
        settings.topology = '../test_data/test.prmtop'
        settings.commit_fwd = [[1, 2], [3, 4], [1.0, 1.5], ['t', 'lt']]
        settings.commit_bwd = [[1, 2], [3, 4], [0.5, 2.0], ['lt', 'gt']]
        with pytest.raises(ValueError):
            utilities.check_commit('../test_data/test.rst7', settings)
        settings.commit_fwd = [[1, 2], [3, 4], [1.0, 1.5], ['gt', 'lt']]
        settings.commit_bwd = [[1, 2], [3, 4], [1.0, 1.5], ['t', 'lt']]
        with pytest.raises(ValueError):
            utilities.check_commit('../test_data/test.rst7', settings)

    def test_get_cvs_no_qdot(self):
        """Tests get_cvs with a dummy coordinate file and include_qdot = False"""
        settings = configure('../../data/atesa.config')
        settings.include_qdot = False
        settings.topology = '../test_data/test.prmtop'
        settings.cvs = ['pytraj.distance(traj, \'@1 @3\')[0]', 'pytraj.distance(traj, \'@2 @4\')[0]']
        result = utilities.get_cvs('../test_data/test.rst7', settings)
        assert len(result.split(' ')) == 2
        assert float(result.split(' ')[0]) == pytest.approx(1.09, 1e-3)
        assert float(result.split(' ')[1]) == pytest.approx(1.69, 1e-3)

    def test_get_cvs_with_qdot_broken(self):
        """Tests get_cvs with a dummy coordinate file, include_qdot = True, and a coordinate file without velocities"""
        settings = configure('../../data/atesa.config')
        settings.include_qdot = True
        settings.topology = '../test_data/test.prmtop'
        settings.cvs = ['pytraj.distance(traj, \'@1 @3\')[0]', 'pytraj.distance(traj, \'@2 @4\')[0]']
        with pytest.raises(IndexError):
            result = utilities.get_cvs('../test_data/test.rst7', settings)

    def test_get_cvs_with_qdot(self):
        """Tests get_cvs with a dummy coordinate file, include_qdot = True, and a coordinate file with velocities"""
        settings = configure('../../data/atesa.config')
        settings.include_qdot = True
        settings.topology = '../test_data/test.prmtop'
        settings.cvs = ['pytraj.distance(traj, \'@1 @3\')[0]', 'pytraj.distance(traj, \'@2 @4\')[0]']
        result = utilities.get_cvs('../test_data/test_velocities.rst7', settings)
        assert len(result.split(' ')) == 4
        assert float(result.split(' ')[0]) == pytest.approx(1.089, 1e-3)
        assert float(result.split(' ')[1]) == pytest.approx(1.728, 1e-3)
        assert float(result.split(' ')[2]) == pytest.approx(-0.252, 1e-2)
        assert float(result.split(' ')[3]) == pytest.approx(0.282, 1e-2)

    def test_get_cvs_no_qdot_reduce(self):
        """Tests get_cvs with a dummy coordinate file, include_qdot = False, and reduce = True"""
        settings = configure('../../data/atesa.config')
        settings.include_qdot = False
        settings.topology = '../test_data/test.prmtop'
        settings.cvs = ['pytraj.distance(traj, \'@1 @3\')[0]', 'pytraj.distance(traj, \'@2 @4\')[0]']
        with pytest.raises(FileNotFoundError):  # no as.out present yet
            result = utilities.get_cvs('../test_data/test.rst7', settings, reduce=True)
        shutil.copy('../test_data/as.out', 'as.out')    # copy in the necessary file
        result = utilities.get_cvs('../test_data/test.rst7', settings, reduce=True)
        assert len(result.split(' ')) == 2
        assert float(result.split(' ')[0]) == pytest.approx(0.571, 1e-2)
        assert float(result.split(' ')[1]) == pytest.approx(0.326, 1e-2)

    def test_rev_vels(self):
        """Tests rev_vels by comparing to a known-correct file"""
        rev_file = utilities.rev_vels('../test_data/test_velocities.rst7')
        assert filecmp.cmp('../test_data/test_velocities_reversed.rst7', rev_file)
        os.remove(rev_file)

    def test_evaluate_rc(self):
        """Tests evaluate_rc"""
        cv_list = [1, 0.3, -3, 1.5002]  # various data types and precisions
        rc_defn = '2*CV1 + 0.3*CV2 + CV3/0.4 - numpy.sin(CV4)'
        assert utilities.evaluate_rc(rc_defn, cv_list) == pytest.approx(-6.408, 1E-2)

    def test_resample(self):
        """Tests resample"""
        settings = configure('../../data/atesa.config')
        with pytest.raises(FileNotFoundError):      # run without making restart.pkl
            utilities.resample(settings)
        settings.degeneracy = 1
        settings.initial_coordinates = ['../test_data/test.rst7', '../test_data/test_two_init.rst7']
        settings.rc_definition = '1.00 + 2.34*CV1 - 5.67*CV2'
        settings.rc_reduced_cvs = False
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]', 'pytraj.angle(traj, \'@2 @3 @4\')[0]']
        settings.include_qdot = False
        settings.topology = '../test_data/test.prmtop'
        allthreads = main.init_threads(settings)
        allthreads[0].history.init_coords = [['../test_data/test_velocities_init.rst7', '../test_data/test_velocities_init_bwd.rst7'],
                                             ['../test_data/test_two_init.rst7', '../test_data/test_two_init_bwd.rst7']]
        allthreads[0].history.prod_results = [['fwd', 'bwd'], ['bwd', 'bwd']]
        allthreads[0].history.prod_trajs = [['../test_data/test.nc', '../test_data/test.nc'], ['../test_data/test.nc', '../test_data/test.nc']]
        allthreads[1].history.init_coords = [['../test_data/test_velocities_init.rst7', '../test_data/test_velocities_init_bwd.rst7'],
                                             ['../test_data/test_two_init.rst7', '../test_data/test_two_init_bwd.rst7']]
        allthreads[1].history.prod_results = [['fwd', 'fwd'], ['bwd', 'fwd']]
        allthreads[1].history.prod_trajs = [['../test_data/test.nc', '../test_data/test.nc'], ['../test_data/test.nc', '../test_data/test.nc']]
        allthreads[0].history.timestamps = [1, 3]
        allthreads[1].history.timestamps = [2, 4]
        pickle.dump(allthreads, open('restart.pkl', 'wb'), protocol=2)  # file now exists
        utilities.resample(settings, full_cvs=True)
        assert os.path.exists('as_raw.out')
        # assert os.path.exists('as_decorr.out')    # todo: reinstitute this test once I have a good restart.pkl file to test with; todo: figure out what I meant by this
        raw_lines = open('as_raw.out', 'r').readlines()
        assert len(raw_lines) == 4
        assert not False in [raw_lines[0][0] == 'B', raw_lines[1][0] == 'A', raw_lines[2][0] == 'B', raw_lines[3][0] == 'A']
        new_lines = open('as_full_cvs.out', 'r').readlines()
        cmp_lines = open('../test_data/as_full_cvs.out', 'r').readlines()
        for line_index in range(len(new_lines)):
            for item_index in range(len(new_lines[line_index].split())):
                assert pytest.approx(float(new_lines[line_index].split()[item_index]), float(cmp_lines[line_index].split()[item_index]), 1e-3)

    def test_interpret_cv(self):
        """Test interpret_cv for a distance, angle, dihedral, and difference of distances for both pytraj and mdtraj"""
        settings = configure('../../data/atesa.config')
        settings.cvs = ['pytraj.distance(traj, \'@1 @2\')[0]',
                        'pytraj.angle(traj, \'@1 @2 @3\')[0]',
                        'pytraj.dihedral(traj, \'@1 @2 @3 @4\')[0]',
                        'pytraj.distance(traj, \'@1 @2\')[0] - pytraj.distance(traj, \'@3 @4\')[0]',
                        'mdtraj.compute_distances(mtraj, numpy.array([[83, 84]]))[0][0] * 10',
                        'mdtraj.compute_angles(mtraj, numpy.array([[83, 85, 86]]))[0][0] * 180 / numpy.pi',
                        'mdtraj.compute_dihedrals(mtraj, numpy.array([[84, 83, 85, 86]]))[0][0] * 180 / numpy.pi',
                        '(mdtraj.compute_distances(mtraj, numpy.array([[101, 103]]))[0][0] * 10) - (mdtraj.compute_distances(mtraj, numpy.array([[101, 104]]))[0][0] * 10)']
        assert utilities.interpret_cv(1, settings) == (['1', '2'], 'distance', 2)
        assert utilities.interpret_cv(2, settings) == (['1', '2', '3'], 'angle', 3)
        assert utilities.interpret_cv(3, settings) == (['1', '2', '3', '4'], 'dihedral', 4)
        assert utilities.interpret_cv(4, settings) == (['1', '2', '3', '4'], 'diffdistance', 4)
        assert utilities.interpret_cv(5, settings) == (['84', '85'], 'distance', 2)
        assert utilities.interpret_cv(6, settings) == (['84', '86', '87'], 'angle', 3)
        assert utilities.interpret_cv(7, settings) == (['85', '84', '86', '87'], 'dihedral', 4)
        assert utilities.interpret_cv(8, settings) == (['102', '104', '102', '105'], 'diffdistance', 4)

    @classmethod
    def teardown_method(self, method):
        "Runs at end of class"
        for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
            os.remove(filename)



"""
Script to evaluate and store information error in support of the information error convergence criterion in aimless
shooting. Implemented as a separate script to facilitate multiprocessing.
"""

import sys
import os
import time
import re
import pickle
import subprocess
import shutil
import glob
import warnings
from atesa import utilities
# from statsmodels.tsa.stattools import kpss    # deprecated

def main():
    """
    Evaluate the information error during aimless shooting and output results to info_err.out.

    Reads decorrelated aimless shooting output files from the present directory to evaluate the mean parametric variance
    based on the Godambe information error of the aimless shooting process.

    This function depends on lmax.py for evaluations of the information error.

    Parameters
    ----------
    None

    Returns
    -------
    None

    """

    # Initialize info_err.out if it does not yet exist
    if not os.path.exists('info_err.out'):
        open('info_err.out', 'w').close()

    # Initialize regular expressions for obtaining strings later
    pattern = re.compile('[0-9.]+')     # pattern to match information error number in lmax output file
    pattern2 = re.compile('[0-9]+')     # pattern to match amount of data in decorrelated datafile names

    # Assemble list of data lengths to perform information error evaluation for
    datalengths = [int(pattern2.findall(item)[-1]) for item in glob.glob('as_decorr_*.out') if not 'lmax' in item]
    if not datalengths:
        raise RuntimeError('attempted to evaluate information error, but found no files matching the pattern '
                           '\'as_decorr_*.out\' in the directory: ' + os.getcwd() + '\n'
                           'Was utilities.resample called first to produce these files?')
    datalengths = sorted(datalengths, key=lambda x: int(x))     # sort ascending to order chronologically
    if datalengths:
        open('info_err_temp.out', 'w').close()

    # Get length of largest datafile to perform two_line_test model optimization on
    length = max(datalengths)
    datalengths.remove(length)  # so as to skip repeat optimization later

    # Read settings from pickle file
    try:
        settings = pickle.load(open('settings.pkl', 'rb'))
    except FileNotFoundError:   # replace with more informative error message
        raise FileNotFoundError('the working directory must contain a valid settings.pkl file, which is generated '
                                'automatically when running ATESA, but one was not found in the working directory: '
                                + os.getcwd())

    # Exit if the datafile has no content (as in, no decorrelated steps (almost impossible, included for robustness))
    if len(open('as_decorr_' + str(length) + '.out', 'r').readlines()) == 0:
        warnings.warn('skipping information error evaluation because decorrelated datafile as_decorr_' + str(length) +
                      '.out is empty. This may indicate an unusual error, especially if you see this message multiple '
                      'times in the course of sampling.', RuntimeWarning)
        return None

    # Run likelihood maximization on largest decorrelated datafile
    if os.path.exists(str(length) + '_lmax.out'):   # remove pre-existing output file if it exists
        os.remove(str(length) + '_lmax.out')
    if settings.include_qdot:
        q_str = 'present'
    else:
        q_str = 'absent'
    command = 'lmax.py -i as_decorr_' + str(length) + '.out -q ' + q_str + ' ' + settings.information_error_lmax_string + ' -o ' + str(length) + '_lmax.out --quiet'
    subprocess.check_call(command.split(' '))     # check_call waits for completion
    if not os.path.exists(str(length) + '_lmax.out'):
        raise FileNotFoundError('Likelihood maximization did not produce output file:' + str(length) + '_lmax.out')

    # Assemble model dimensions just obtained for calling lmax for other data sets
    model_str = open(str(length) + '_lmax.out', 'r').readlines()[1]
    model_str = model_str[model_str.rindex(': ') + 2:]
    dims = ''
    while 'CV' in model_str:
        model_str = model_str[model_str.index('CV') + 2:]    # chop off everything up to and including the next 'CV'
        if ' ' in model_str:
            dims += model_str[:model_str.index(' ')] + ' '  # CV index is everything up until the next space
        else:
            dims += model_str[:model_str.index('\n')]       # last dimensions has a newline instead of a space
    if dims == '':
        raise RuntimeError('Likelihood maximization output file is improperly formatted: ' + str(length) + '_lmax.out')

    # Call lmax for each further dataset and write new info_err output file
    for datalength in datalengths:
        command = 'lmax.py -i as_decorr_' + str(datalength) + '.out -q ' + q_str + ' -f ' + dims + ' -k ' + str(int(len(dims.split(' ')))) + ' -o ' + str(datalength) + '_lmax.out --quiet'
        subprocess.check_call(command.split(' '))
        inf_err = float(pattern.findall(open(str(datalength) + '_lmax.out', 'r').readlines()[-1])[0])
        open('info_err_temp.out', 'a').write(str(datalength) + ' ' + str(inf_err) + '\n')
        open('info_err_temp.out', 'a').close()

    # Add information error from previously completed lmax for this length
    inf_err = float(pattern.findall(open(str(length) + '_lmax.out', 'r').readlines()[-1])[0])
    open('info_err_temp.out', 'a').write(str(length) + ' ' + str(inf_err) + '\n')
    open('info_err_temp.out', 'a').close()

    # Move info_err_temp.out to info_err.out
    shutil.move('info_err_temp.out', 'info_err.out')

    # Deprecated (no longer using KPSS)
    # # Next, evaluate the KPSS statistic for the dataset in info_err.out
    # # Begin by suppressing warnings that occur frequently during normal operation of kpss
    # warnings.filterwarnings('ignore', message='p-value is greater than the indicated p-value')
    # warnings.filterwarnings('ignore', message='invalid value encountered in double_scalars')
    # warnings.filterwarnings('ignore', message='divide by zero encountered in double_scalars')
    #
    # # Get data back out from info_err.out (weird but robust)
    # info_err_lines = open('info_err.out', 'r').readlines()
    # info_errs = [float(line.split(' ')[1]) for line in info_err_lines]  # cast to float removes '\n'
    # kpssresults = []
    # for cut in range(len(info_errs) - 1):
    #     try:
    #         kpssresult = (kpss(info_errs[0:cut + 1], lags='auto')[1] - 0.01) / (0.1 - 0.01)
    #     except (ValueError, OverflowError):
    #         kpssresult = 1  # 100% certainty of non-convergence!
    #     kpssresults.append(kpssresult)
    #
    # # Write new info_err.out with kpss values
    # open('info_err.out', 'w').write(info_err_lines[0])  # no KPSS statistic for the first line
    # for line_index in range(1, len(info_err_lines)):
    #     line_data = [float(item) for item in info_err_lines[line_index].split(' ')]
    #     new_line = '%.0f' % line_data[0] + ' ' + '%.3f' % line_data[1] + ' ' + '%.3f' % kpssresults[line_index - 1] + '\n'
    #     open('info_err.out', 'a').write(new_line)
    # open('info_err.out', 'a').close()


if __name__ == "__main__":
    main()



"""
Interface for TaskManager objects. New TaskManagers can be implemented by constructing a new class that inherits from
TaskManager and implements its abstract methods.
"""

import abc
import subprocess
import re
import time

class TaskManager(abc.ABC):
    """
    Abstract base class for task managers.

    Implements methods for all of the task manager-specific tasks that ATESA might need.

    """

    @abc.abstractmethod
    def submit_batch(self, filename, settings):
        """
        Submit a batch file to the task manager.

        Parameters
        ----------
        filename : str
            Name of batch file to submit
        settings : argparse.Namespace
            Settings namespace object

        Returns
        -------
        jobid : str
            Identification number for this task, such that it can be cancelled by referring to this string

        """

        pass


class AdaptSimple(TaskManager):
    """
    Adapter class for my simple, no-frills task manager.

    Just interfaces directly with the batch system through the terminal.

    """

    def submit_batch(self, filename, settings):
        try:    # import here to avoid circular import
            import factory
        except ModuleNotFoundError:
            import atesa.factory as factory

        batchsystem = factory.batchsystem_factory(settings.batch_system)
        command = batchsystem.get_submit_command().replace('{file}', filename)

        if settings.DEBUG:
            command = 'echo "this is a nonsense string for testing purposes: 123456, now here are some garbage symbols: ?!@#$/\';:[]+=_-.<,>"'

        count = 1
        max_tries = 5
        output = 'first_attempt'
        errors = ['first_attempt', 'slurm_load_jobs', 'slurm_receive_msg', 'send/recv']  # error messages to retry on
        while True in [error in output for error in errors] and count <= max_tries:
            if not output == 'first_attempt':
                time.sleep(30)      # wait 30 seconds before trying again
                count += 1
            process = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                       close_fds=True, shell=True)
            output = process.stdout.read().decode()

        # Use a regular expression to extract the jobid from this string
        pattern = re.compile('[0-9]+')  # todo: it's not inconceivable that this should fail in some cases. Consider moving building this pattern to a method of BatchSystem.
        try:
            return re.findall(pattern, output)[0]
        except IndexError:  # no number in the output
            raise RuntimeError('unable to submit batch job: ' + filename + '\nMessage from batch system: ' + output)




"""
Unit and regression test for boltzmann_weight.py.
"""

# Import package, test suite, and other packages as needed
import atesa
import sys
import os
import shutil
import pytest
import pickle
import subprocess
import glob
from atesa import boltzmann_weight
from atesa.configure import configure

class Tests(object):
    def setup_method(self, test_method):
        try:
            if not os.path.exists('atesa/tests/test_temp'):
                os.mkdir('atesa/tests/test_temp')
            os.chdir('atesa/tests/test_temp')
        except FileNotFoundError:
            pass

    def test_main(self):
        """Tests boltzmann_weight using example EPS output in test_data"""
        kwargs = {'i': '../test_data/eps_example.out', 'o': 'fep_test.out', 't': 350, 'n': 4, 'b': -1, 'c': 1, 'noplot': True, 'e': 0.8, 'slope_only': False}
        # There's a necessary random component here that can potentially cause a RuntimError to be raised, but the odds
        # of it failing three times in a row without there actually being something wrong are extremely small, so this
        # is the best I can think to do...
        try:
            boltzmann_weight.main(**kwargs)
        except RuntimeError:
            try:
                boltzmann_weight.main(**kwargs)
            except RuntimeError:
                boltzmann_weight.main(**kwargs)
        compare_lines = open('../test_data/fep_example.out', 'r').readlines()
        line_index = 0
        for line in open('fep_test.out', 'r').readlines():
            # Compare first and second entries in each line (third is bootstrapped and so will vary)
            assert pytest.approx(float(line.split()[0])) == float(compare_lines[line_index].split()[0])
            assert pytest.approx(float(line.split()[1])) == float(compare_lines[line_index].split()[1])
            line_index += 1

    @classmethod
    def teardown_method(self, method):
        "Runs at end of class"
        for filename in glob.glob(sys.path[0] + '/atesa/tests/test_temp/*'):
            os.remove(filename)
